#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import datetime
import io
import json
import logging
import os.path
import zipfile
from contextlib import contextmanager
from typing import Union, Optional, Tuple, Any, Dict, Set, Iterable

from ..importer import (
    BaseImporter, Record, Folder, RecordField, RecordReferences, SharedFolder, Attachment
)
from ... import utils
from ...params import KeeperParams
from ...record_types import FieldTypes


class OnePasswordImporter(BaseImporter):
    def __init__(self):
        super(OnePasswordImporter, self).__init__()

    def extension(self):
        return '1pux'

    def support_folder_filter(self):
        return True

    @staticmethod
    def get_field_value(field):  # type: (dict) -> Optional[Tuple[str, Any]]
        if isinstance(field, dict):
            if 'value' in field:
                value = field['value']
                if isinstance(value, dict):
                    for key in value:
                        return key, value[key]

    def do_import(self, filename, **kwargs):
        # type: (BaseImporter, str, dict) -> Iterable[Union[Record, SharedFolder]]
        record_types = {}
        params = kwargs.get('params')
        if isinstance(params, KeeperParams):
            if params.record_type_cache:
                for rts in params.record_type_cache.values():
                    try:
                        rto = json.loads(rts)
                        if '$id' in rto and 'fields' in rto:
                            record_types[rto['$id']] = rto['fields']
                    except:
                        pass

        if not os.path.isfile(filename):
            fn = os.path.expanduser(filename)
            if os.path.isfile(fn):
                filename = fn

        with zipfile.ZipFile(filename, mode='r') as zip_file:
            files = set(zip_file.namelist())
            if 'export.data' not in files:
                return
            with zip_file.open('export.data', mode='r') as data_file:
                export_data = json.load(data_file)

            filter_folder = kwargs.get('filter_folder')
            if 'accounts' in export_data:
                for account in export_data['accounts']:
                    for vault in account.get('vaults', []):
                        vault_name = None
                        is_shared = False
                        if 'attrs' in vault:
                            name = vault['attrs'].get('name')
                            if filter_folder and name != filter_folder:
                                continue

                            vault_type = vault['attrs'].get('type', '')
                            if vault_type == 'E':
                                vault_name = name
                                is_shared = True
                                sf = SharedFolder()
                                sf.path = vault_name
                                yield sf
                            elif vault_type == 'U':
                                vault_name = name

                        vault_records = {}
                        references = {}   # type: Dict[str, Set[str]]
                        for item in vault.get('items', []):
                            if item.get('trashed') is True:
                                continue
                            record = Record()
                            record.uid = item.get('uuid') or utils.generate_uid()
                            lm = item.get('updatedAt')
                            if isinstance(lm, int) and lm > 0:
                                record.last_modified = lm
                            category = item.get('categoryUuid', '')
                            if category in ('001', '005', '110', '112'):
                                record.type = 'login'
                            elif category == '002':
                                record.type = 'bankCard'
                            elif category == '003':
                                record.type = 'encryptedNotes'
                            elif category == '004':
                                record.type = 'contact'
                            elif category == '006':
                                record.type = 'file'
                            elif category == '100':
                                record.type = 'softwareLicense'
                            elif category == '101':
                                record.type = 'bankAccount'
                            elif category == '102':
                                record.type = 'databaseCredentials'
                            elif category == '103':
                                record.type = 'driverLicense'
                            elif category == '104':
                                record.type = 'membership'
                            elif category == '105':
                                record.type = 'membership'
                            elif category == '106':
                                record.type = 'passport'
                            elif category == '107':
                                record.type = 'membership'
                            elif category == '108':
                                record.type = 'ssnCard'
                            elif category == '109':
                                record.type = 'wifiCredentials'
                            elif category == '111':
                                record.type = 'serverCredentials'
                            elif category == '113':
                                record.type = 'encryptedNotes'
                            elif category == '114':
                                record.type = 'sshKeys'
                            elif category == '115':
                                record.type = 'login'

                            if 'overview' in item:
                                overview = item['overview']
                                record.title = overview['title']
                                if 'urls' in overview:
                                    for url_obj in overview['urls']:
                                        label = url_obj.get('label', '')
                                        url = url_obj.get('url', '')
                                        if url:
                                            if record.login_url:
                                                record.fields.append(RecordField(type='url', label=label, value=url))
                                            else:
                                                record.login_url = url
                                if 'tags' in overview:
                                    record.folders = []
                                    for tag in overview['tags']:    # type: str
                                        tag = tag.replace('/', '\\')
                                        folder = Folder()
                                        if vault_name:
                                            if is_shared:
                                                folder.domain = vault_name
                                            else:
                                                folder.path = vault_name
                                        if folder.path:
                                            folder.path += '\\'
                                        else:
                                            folder.path = ''
                                        folder.path += tag
                                        record.folders.append(folder)
                                if not record.folders and vault_name:
                                    record.folders = []
                                    folder = Folder()
                                    if vault_name:
                                        if is_shared:
                                            folder.domain = vault_name
                                        else:
                                            folder.path = vault_name
                                    record.folders.append(folder)

                            details = item.get('details')
                            if not details:
                                continue
                            if 'loginFields' in details:
                                for login_field in details['loginFields']:
                                    value = login_field.get('value')
                                    if not value:
                                        continue
                                    designation = login_field.get('designation')
                                    if designation == 'username':
                                        record.login = value
                                    elif designation == 'password':
                                        record.password = value
                            if 'password' in details:
                                password = details['password']
                                if password:
                                    if record.password:
                                        record.fields.append(RecordField(type='secret', label='Password', value=password))
                                    else:
                                        record.password = password
                            if 'notesPlain' in details:
                                record.notes = details.get('notesPlain', '')
                            if 'documentAttributes' in details:
                                field_value = details['documentAttributes']
                                if isinstance(field_value, dict):
                                    file_name = field_value.get('fileName')
                                    document_id = field_value.get('documentId')
                                    if file_name and document_id:
                                        if record.attachments is None:
                                            record.attachments = []
                                        size = field_value.get('decryptedSize')
                                        zip_path = f'files/{document_id}__{file_name}'
                                        record.attachments.append(
                                            OnePasswordAttachment(filename, zip_path, file_name, size))
                            if 'sections' in details:
                                rt = record_types.get(record.type, [])
                                for section in details['sections']:
                                    section_label = section.get('title', '')
                                    section_type = section.get('name')
                                    fields = section.get('fields', [])
                                    if not isinstance(fields, list):
                                        continue
                                    field_dict = {x['id']: x for x in fields if 'id' in x}
                                    if section_type == 'security questions':
                                        qs = set()
                                        questions = []
                                        for field_id in field_dict:
                                            field = field_dict[field_id]
                                            if 'value' in field:
                                                field_value = field['value']
                                                if 'concealed' in field_value:
                                                    qa = FieldTypes['securityQuestion'].value.copy()
                                                    qa['question'] = field.get('title', '')
                                                    qa['answer'] = field_value.get('concealed', '')
                                                    questions.append(qa)
                                                    qs.add(field_id)
                                        for field_id in qs:
                                            del field_dict[field_id]
                                        if questions:
                                            record.fields.append(RecordField(type='securityQuestion', label=section_label, value=questions))
                                    if 'hostname' in field_dict:
                                        host = FieldTypes['host'].value.copy()
                                        hostname_field = field_dict.pop('hostname', None)
                                        hostname = OnePasswordImporter.get_field_value(hostname_field)
                                        if hostname and hostname[1]:
                                            host['hostName'] = hostname[1]
                                        port = OnePasswordImporter.get_field_value(field_dict.pop('port', None))
                                        if port and port[1]:
                                            host['port'] = port[1]
                                        field_label = hostname_field.get('title', section_label)
                                        field_label = OnePasswordImporter.adjust_field_label(record, 'host', field_label, rt)
                                        record.fields.append(RecordField(type='host', label=field_label, value=host))
                                    if 'pop_server' in field_dict or 'smtp_server' in field_dict:
                                        for prefix in ('pop', 'smtp'):
                                            type_value = OnePasswordImporter.get_field_value(field_dict.pop(f'{prefix}_type', None))
                                            if type_value and type_value[1]:
                                                account_type = type_value[1].upper()
                                            else:
                                                account_type = prefix.upper()
                                            username_value = OnePasswordImporter.get_field_value(field_dict.pop(f'{prefix}_username', None))
                                            password_value = OnePasswordImporter.get_field_value(field_dict.pop(f'{prefix}_password', None))
                                            server_value = OnePasswordImporter.get_field_value(field_dict.pop(f'{prefix}_server', None))
                                            port_value = OnePasswordImporter.get_field_value(field_dict.pop(f'{prefix}_port', None))
                                            security_value = OnePasswordImporter.get_field_value(field_dict.pop(f'{prefix}_security', None))
                                            authentication_value = OnePasswordImporter.get_field_value(field_dict.pop(f'{prefix}_authentication', None))
                                            username = ''
                                            password = ''
                                            host = None
                                            if username_value:
                                                username = username_value[1]
                                            if password_value:
                                                password = password_value[1]
                                            if server_value and server_value[1]:
                                                host = FieldTypes['host'].value.copy()
                                                host['hostName'] = server_value[1]
                                                if port_value:
                                                    host['port'] = port_value[1]
                                            if username or password or host:
                                                if record.type == 'serverCredentials' and not record.login and not record.password:
                                                    record.login = username
                                                    record.password = password
                                                    if host:
                                                        record.fields.append(RecordField(type='host', label='', value=host))
                                                else:
                                                    if username:
                                                        record.fields.append(RecordField(type='login', label=f'{account_type} Username', value=username))
                                                    if password:
                                                        record.fields.append(RecordField(type='secret', label=f'{account_type} Password', value=password))
                                                    if host:
                                                        record.fields.append(RecordField(type='host', label=f'{account_type} Server', value=host))
                                                if security_value and security_value[1]:
                                                    record.fields.append(RecordField(type='text', label=f'{account_type} Security', value=security_value[1]))
                                                if authentication_value and authentication_value[1]:
                                                    record.fields.append(RecordField(type='text', label=f'{account_type} Auth Method', value=authentication_value[1]))
                                    if 'reminderq' in field_dict and 'remindera' in field_dict:
                                        q = OnePasswordImporter.get_field_value(field_dict.pop('reminderq', None))
                                        a = OnePasswordImporter.get_field_value(field_dict.pop('remindera', None))
                                        if q and q[1]:
                                            qa = FieldTypes['securityQuestion'].value.copy()
                                            qa['question'] = q[1]
                                            if a and a[1]:
                                                qa['answer'] = a[1]
                                            record.fields.append(RecordField(type='securityQuestion', label='Reminder', value=qa))
                                    if 'firstname' in field_dict and 'lastname' in field_dict:   # name field
                                        name = FieldTypes['name'].value.copy()
                                        firstname = OnePasswordImporter.get_field_value(field_dict.pop('firstname', None))
                                        if firstname and firstname[1]:
                                            name['first'] = firstname[1]
                                        lastname = OnePasswordImporter.get_field_value(field_dict.pop('lastname', None))
                                        if lastname and lastname[1]:
                                            name['last'] = lastname[1]
                                        initial = OnePasswordImporter.get_field_value(field_dict.pop('initial', None))
                                        if lastname and lastname[1]:
                                            name['middle'] = initial[1]
                                        field_label = OnePasswordImporter.adjust_field_label(record, 'name', section_label, rt)
                                        record.fields.append(RecordField(type='name', label=field_label, value=name))
                                    if 'ccnum' in field_dict and 'expiry' in field_dict:   # credit card field
                                        cardholder = OnePasswordImporter.get_field_value(field_dict.pop('cardholder', None))
                                        if cardholder and cardholder[1]:
                                            record.fields.append(RecordField(type='text', label='cardholderName', value=cardholder[1]))
                                        card = FieldTypes['paymentCard'].value.copy()
                                        ccnum = OnePasswordImporter.get_field_value(field_dict.pop('ccnum', None))
                                        if ccnum and ccnum[1]:
                                            card['cardNumber'] = ccnum[1]
                                        card_type = OnePasswordImporter.get_field_value(field_dict.pop('type', None))
                                        validFrom = OnePasswordImporter.get_field_value(field_dict.pop('validFrom', None))
                                        cvv = OnePasswordImporter.get_field_value(field_dict.pop('cvv', None))
                                        if cvv and cvv[1]:
                                            card['cardSecurityCode'] = cvv[1]
                                        expiry = OnePasswordImporter.get_field_value(field_dict.pop('expiry', None))
                                        if expiry and expiry[0] == 'monthYear' and isinstance(expiry[1], int):
                                            if expiry[1] > 0:
                                                card['cardExpirationDate'] = f'{expiry[1]%100:02}/{expiry[1]//100}'
                                        field_label = OnePasswordImporter.adjust_field_label(record, 'paymentCard', section_label, rt)
                                        record.fields.append(RecordField(type='paymentCard', label=field_label, value=card))
                                    if 'accountNo' in field_dict and 'routingNo' in field_dict:
                                        bankAccount = FieldTypes['bankAccount'].value.copy()
                                        accountType = OnePasswordImporter.get_field_value(field_dict.pop('accountType', None))
                                        if accountType and accountType[1]:
                                            bankAccount['accountType'] = str(accountType[1]).capitalize()
                                        accountNo = OnePasswordImporter.get_field_value(field_dict.pop('accountNo', None))
                                        if accountNo and accountNo[1]:
                                            bankAccount['accountNumber'] = accountNo[1]
                                        routingNo = OnePasswordImporter.get_field_value(field_dict.pop('routingNo', None))
                                        if routingNo and routingNo[1]:
                                            bankAccount['routingNumber'] = routingNo[1]
                                        field_label = OnePasswordImporter.adjust_field_label(record, 'bankAccount', section_label, rt)
                                        record.fields.append(RecordField(type='bankAccount', label=field_label, value=bankAccount))
                                    if len(field_dict) > 0:
                                        for field_id in field_dict:
                                            field = field_dict[field_id]
                                            value_pair = OnePasswordImporter.get_field_value(field)
                                            if not value_pair:
                                                continue
                                            field_type, field_value = value_pair
                                            if not field_value:
                                                continue
                                            ft = 'text'
                                            fl = field.get('title', '')
                                            if record.type == 'wifiCredentials' and field_id == 'name':
                                                fl = 'SSID'
                                                fv = field_value
                                            elif field_id in ('name', 'owner', 'fullname', 'member_name'):
                                                last, sep, rest = field_value.partition(',')
                                                if sep == ',':
                                                    names = rest.split(' ')
                                                    names.append(last)
                                                else:
                                                    names = field_value.split(' ')
                                                names = [x.strip() for x in names]
                                                names = [x for x in names if x]

                                                ft = 'name'
                                                fv = FieldTypes['name'].value.copy()
                                                if len(names) > 0:
                                                    fv['last'] = names.pop()
                                                if len(names) > 0:
                                                    fv['first'] = names.pop(0)
                                                if len(names) > 0:
                                                    fv['middle'] = ' '.join(names)
                                                fl = OnePasswordImporter.adjust_field_label(record, 'name', fl, rt)
                                            elif field_id.endswith('website'):
                                                ft = 'url'
                                                fv = field_value
                                            elif field_id in ('number', 'membership_no', 'reg_code'):
                                                if record.type in ('ssnCard', 'membership', 'driverLicense', 'healthInsurance', 'passport'):
                                                    ft = 'accountNumber'
                                                elif record.type == 'softwareLicense':
                                                    ft = 'licenseNumber'
                                                fv = field_value
                                                fl = OnePasswordImporter.adjust_field_label(record, ft, fl, rt)
                                            elif field_id == 'url':
                                                if record.login_url:
                                                    ft = 'url'
                                                    fv = field_value
                                                else:
                                                    record.login_url = field_value
                                                    continue
                                            elif field_id == 'email':
                                                ft = 'email'
                                                fv = field_value
                                                fl = OnePasswordImporter.adjust_field_label(record, ft, fl, rt)
                                            elif field_id == 'reason' and record.type == 'encryptedNotes':
                                                ft = 'note'
                                                fv = field_value
                                                fl = ''
                                            elif field_id == 'username':
                                                if record.login:
                                                    ft = 'email' if field_value.find('@') > 0 else 'login'
                                                else:
                                                    record.login = field_value
                                                    continue
                                            elif field_type == 'reference':
                                                ref_uid = field_value
                                                if record.uid not in references:
                                                    references[record.uid] = set()
                                                references[record.uid].add(ref_uid)
                                                continue
                                            elif field_type == 'concealed':
                                                if field_id.endswith('password') or \
                                                        field_id == 'credential' or \
                                                        (field_id == 'pin' and record.type == 'membership'):
                                                    if record.password:
                                                        ft = 'secret'
                                                    else:
                                                        record.password = field_value
                                                        continue
                                                elif record.type == 'sshKeys' and not record.password:
                                                    record.password = field_value
                                                    continue
                                                else:
                                                    ft = 'pinCode'
                                                    if record.type == 'bankCard':
                                                        fl = ''
                                                fv = field_value
                                            elif field_type == 'phone':
                                                ft = 'phone'
                                                fv = FieldTypes['phone'].value.copy()
                                                fv['number'] = field_value
                                                if field_id.startswith('cell'):
                                                    fv['type'] = 'Mobile'
                                                elif field_id.startswith('home'):
                                                    fv['type'] = 'Home'
                                                elif field_id.startswith('bus'):
                                                    fv['type'] = 'Business'
                                                fl = OnePasswordImporter.adjust_field_label(record, ft, fl, rt)
                                            elif field_type == 'url':
                                                if record.login_url:
                                                    ft = 'url'
                                                    fv = field_value
                                                else:
                                                    record.login_url = field_value
                                                    continue
                                            elif field_type == 'email':
                                                ft = 'email'
                                                if isinstance(field_value, dict):
                                                    fv = field_value.get('email_address')
                                                else:
                                                    logging.debug('\"email\" field type expected to be \"dict\"')
                                                    continue
                                            elif field_type in ('date', 'monthYear'):
                                                if isinstance(field_value, int):
                                                    if field_type == 'monthYear':
                                                        month = field_value % 100
                                                        year = field_value // 100
                                                        d = datetime.date(year, month, 1)
                                                        dt = datetime.datetime.combine(d, datetime.time(), tzinfo=datetime.timezone.utc)
                                                        field_value = int(dt.timestamp())
                                                    fv = field_value * 1000
                                                    if field_id == 'birthdate':
                                                        ft = 'birthDate'
                                                        fl = ''
                                                    elif field_id == 'expiry_date':
                                                        ft = 'expirationDate'
                                                        fl = ''
                                                    elif field_id == 'issue_date':
                                                        ft = 'date'
                                                        fl = 'dateIssued'
                                                    elif field_id == 'order_date':
                                                        ft = 'date'
                                                        fl = 'dateActive'
                                                    else:
                                                        ft = 'date'
                                                else:
                                                    continue
                                            elif field_type == 'address' and isinstance(field_value, dict):
                                                ft = 'address'
                                                fv = FieldTypes['address'].value.copy()
                                                fv['street1'] = field_value.get('street', '')
                                                fv['city'] = field_value.get('city', '')
                                                fv['state'] = field_value.get('state', '')
                                                fv['zip'] = field_value.get('zip', '')
                                                # fv['country'] = field_value.get('country', '')
                                                fl = OnePasswordImporter.adjust_field_label(record, ft, fl, rt)
                                            elif field_type == 'file' and isinstance(field_value, dict):
                                                if record.attachments is None:
                                                    record.attachments = []
                                                file_name = field_value.get('fileName')
                                                document_id = field_value.get('documentId')
                                                if file_name and document_id:
                                                    size = field_value.get('decryptedSize')
                                                    zip_path = f'files/{document_id}__{file_name}'
                                                    record.attachments.append(
                                                        OnePasswordAttachment(filename, zip_path, file_name, size))
                                                continue
                                            elif field_type == 'totp' and isinstance(field_value, str):
                                                if field_value.startswith('otpauth://'):
                                                    fv = field_value
                                                else:
                                                    fv = f'otpauth://totp/?secret={field_value}'
                                                ft = 'oneTimeCode'
                                                fl = OnePasswordImporter.adjust_field_label(record, ft, fl, rt)
                                            elif field_type == 'sshKey' and isinstance(field_value, dict):
                                                ft = 'keyPair'
                                                fv = FieldTypes['privateKey'].value.copy()
                                                fv['privateKey'] = field_value.get('privateKey')
                                                if 'metadata' in field_value:
                                                    fv['publicKey'] = field_value['metadata'].get('publicKey')
                                                fl = OnePasswordImporter.adjust_field_label(record, ft, fl, rt)
                                            elif isinstance(field_value, str):
                                                ft = 'text'
                                                fv = field_value
                                            else:
                                                logging.debug('Unsupported 1Password field type: %s', field_type)
                                                continue
                                            record.fields.append(RecordField(type=ft, label=fl, value=fv))

                            vault_records[record.uid] = record

                        if len(references) > 0:
                            for record_uid in references:
                                record = vault_records.get(record_uid)
                                refs = references[record_uid]
                                if record and isinstance(refs, set):
                                    cards = RecordReferences('card')
                                    addresses = RecordReferences('address')
                                    record.references = []
                                    for ref_uid in refs:
                                        ref_record = vault_records.get(ref_uid)
                                        if ref_record:
                                            if ref_record.type == 'bankCard':
                                                cards.uids.append(ref_uid)
                                            elif ref_record.type == 'address':
                                                addresses.uids.append(ref_uid)
                                    if len(addresses.uids) > 0:
                                        record.references.append(addresses)
                                    if len(cards.uids) > 0:
                                        record.references.append(cards)

                            references.clear()
                        for record in vault_records.values():
                            yield record
                        vault_records.clear()


class OnePasswordAttachment(Attachment):
    def __init__(self, zip_file, zip_path, name, size):
        super().__init__()
        self.zip_file = zip_file
        self.file_uid = zip_path
        self.name = name
        self.size = size

    @contextmanager
    def open(self):  # type: () -> io.BufferedIOBase
        with zipfile.ZipFile(self.zip_file, mode='r') as zip_file:
            yield zip_file.open(self.file_uid, mode='r')












from cryptography.hazmat.primitives.ciphers.aead import AESGCM
import base64
import os
from typing import Optional, Union


def encrypt_aes(data: bytes, key: bytes, iv: bytes = None) -> bytes:
    aesgcm = AESGCM(key)
    iv = iv or os.urandom(12)
    enc = aesgcm.encrypt(iv, data, None)
    return iv + enc


def decrypt_aes(data: bytes, key: bytes) -> bytes:
    aesgcm = AESGCM(key)
    return aesgcm.decrypt(data[:12], data[12:], None)


def bytes_to_base64(b: Union[str, bytes]) -> str:

    if isinstance(b, str) is True:
        b = b.encode()

    return base64.b64encode(b).decode()


def urlsafe_str_to_bytes(s: str) -> bytes:
    b = base64.urlsafe_b64decode(s + '==')
    return b


def str_to_bytes(s: str) -> bytes:
    b = base64.b64decode(s + '==')
    return b


def bytes_to_urlsafe_str(b: Union[str,bytes]) -> str:
    """
    Convert bytes to a URL-safe base64 encoded string.

    Args:
        b (bytes): The bytes to be encoded.

    Returns:
        str: The URL-safe base64 encoded representation of the input bytes.
    """
    if isinstance(b, str) is True:
        b = b.encode()

    return base64.urlsafe_b64encode(b).decode().rstrip('=')


def bytes_to_str(b: bytes) -> str:
    """
    Convert bytes to a URL-safe base64 encoded string.

    Args:
        b (bytes): The bytes to be encoded.

    Returns:
        str: The URL-safe base64 encoded representation of the input bytes.
    """
    return base64.b64encode(b).decode().rstrip('=')


def generate_random_bytes(length: int) -> bytes:
    return os.urandom(length)


def generate_uid_bytes(length: int = 16) -> bytes:
    return generate_random_bytes(length)


def generate_uid_str(uid_bytes: Optional[bytes] = None) -> str:
    if uid_bytes is None:
        uid_bytes = generate_uid_bytes()
    return bytes_to_urlsafe_str(uid_bytes)



from urllib.parse import urlsplit


def get_hostname(server):
    """Split hostname from server URL

    Also removes the port from the hostname if there is one.
    server(str): URL for server
    Returns the hostname component of the URL or None if the server URL evaluates to False
    """
    if server:
        parts = urlsplit(server)
        host = parts[1]
        cp = host.rfind(':')
        if cp > 0:
            host = host[:cp]
        return host.lower()
    else:
        return None


def get_environment(hostname):
    """Get whether dev or qa environment from Keeper server hostname

    hostname(str): The hostname component of the Keeper server URL
    Returns one of 'DEV', 'QA', or None
    """
    environment = None
    if hostname:
        if hostname.startswith('dev.'):
            environment = 'DEV'
        elif hostname.startswith('qa.'):
            environment = 'QA'
    return environment


def get_data_center(hostname):
    """Guess data center from Keeper server hostname

    hostname(str): The hostname component of the Keeper server URL
    Returns one of "EU", "US", "US GOV", or "AU"
    """
    if hostname.endswith('.eu'):
        data_center = 'EU'
    elif hostname.endswith('.com'):
        data_center = 'US'
    elif hostname.endswith('govcloud.keepersecurity.us'):
        data_center = 'US GOV'
    elif hostname.endswith('.au'):
        data_center = 'AU'
    else:
        # Ideally we should determine TLD which might require additional lib
        data_center = hostname
    return data_center



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import os
import logging
import signal
import psutil

from ... import utils
from ...service.config.service_config import ServiceConfig
from ..decorators.logging import logger, debug_decorator
from .process_info import ProcessInfo
from .terminal_handler import TerminalHandler
from .signal_handler import SignalHandler
import sys, subprocess

class ServiceManager:
    """Manages the lifecycle of the service including start, stop, and status operations."""
    
    _flask_app = None
    _is_running = False
    
    @classmethod
    @debug_decorator
    def _handle_shutdown(cls) -> None:
        """Handle service shutdown."""
        logger.debug("Handling service shutdown")
        cls._is_running = False
        cls._flask_app = None
        ProcessInfo.clear()

    @staticmethod
    def get_ssl_context(config_data):
        """
        Get SSL context from configuration data if certificates are available.
        Returns None if no valid certificates are found.
        """
        ssl_context = None
        if config_data.get("certfile") and config_data.get("certpassword"):
            certfile = utils.get_default_path() / os.path.basename(config_data.get("certfile"))
            certpassword = utils.get_default_path() / os.path.basename(config_data.get("certpassword"))
            
            if os.path.exists(certfile) and os.path.exists(certpassword):
                logger.debug('Using SSL certificates')
                ssl_context = (certfile, certpassword)
        
        return ssl_context

    
    @classmethod
    def start_service(cls) -> None:
        """Start the service if not already running."""
        process_info = ProcessInfo.load()
        
        if process_info.pid and process_info.is_running:
            print(f"Error: Commander Service is already running (PID: {process_info.pid})")
            return
            
        SignalHandler.setup_signal_handlers(cls._handle_shutdown)
            
        try:
            service_config = ServiceConfig()
            config_data = service_config.load_config()
            
            if not (port := config_data.get("port")):
                print("Error: Service configuration is incomplete. Please configure the service port in service_config")
                return
            
            from ..config.ngrok_config import NgrokConfigurator
            from ..config.cloudflare_config import CloudflareConfigurator
            
            is_running = True
            queue_enabled = config_data.get("queue_enabled", "y")
            api_version = "v2" if queue_enabled == "y" else "v1"
            
            # Check if SSL is configured to determine the correct protocol
            ssl_context = cls.get_ssl_context(config_data)
            protocol = "https" if ssl_context else "http"
            
            print(f"Commander Service starting on {protocol}://localhost:{port}/api/{api_version}/")
            
            ngrok_pid = NgrokConfigurator.configure_ngrok(config_data, service_config)
            cloudflare_pid = None

            try:
                cloudflare_pid = CloudflareConfigurator.configure_cloudflare(config_data, service_config)
            except Exception as e:
                if ngrok_pid and psutil:
                    try:
                        process = psutil.Process(ngrok_pid)
                        process.terminate()
                        logger.debug(f"Terminated ngrok process {ngrok_pid}")
                    except (psutil.NoSuchProcess, psutil.AccessDenied, OSError) as ngrok_error:
                        logger.debug(f"Error terminating ngrok process: {type(ngrok_error).__name__}")
                elif ngrok_pid:
                    logger.warning("Cannot terminate ngrok process: psutil not available")

                ProcessInfo.clear()

                logger.info(f"\n{str(e)}")
                return

            # Custom logging filter to replace SSL handshake errors with user-friendly message
            class SSLHandshakeFilter(logging.Filter):
                def filter(self, record):
                    # Replace "Bad request version" errors with a clearer message
                    if hasattr(record, 'getMessage'):
                        message = record.getMessage()
                        if "Bad request version" in message and any(ord(c) > 127 for c in message):
                            # Replace the ugly SSL handshake error with a user-friendly message
                            record.msg = "HTTPS request received but HTTPS protocol is not enabled on this service"
                            record.args = ()
                    return True
            
            werkzeug_logger = logging.getLogger('werkzeug')
            werkzeug_logger.setLevel(logging.WARNING)
            werkzeug_logger.addFilter(SSLHandshakeFilter())

            if config_data.get("run_mode") == "background":

                base_dir = os.path.dirname(os.path.abspath(__file__))
                service_module = "keepercommander.service.core.service_app"  # Use module path instead of file path
                python_executable = sys.executable

                # Create logs directory for subprocess output
                log_dir = os.path.join(base_dir, "logs")
                os.makedirs(log_dir, exist_ok=True)
                log_file = os.path.join(log_dir, "service_subprocess.log")

                try:
                    if sys.platform == "win32":
                        subprocess.DETACHED_PROCESS = 0x00000008
                        with open(log_file, 'w') as log_f:
                            cls = subprocess.Popen(
                                [python_executable, '-m', service_module],
                                creationflags=subprocess.DETACHED_PROCESS | subprocess.CREATE_NEW_PROCESS_GROUP,
                                stdout=log_f,
                                stderr=subprocess.STDOUT,  # Combine stderr with stdout
                                cwd=os.getcwd(),  # Use current working directory to access config files
                                env=os.environ.copy()  # Inherit environment variables
                            )
                    else:
                        # For macOS and Linux - improved subprocess handling
                        with open(log_file, 'w') as log_f:
                            cls = subprocess.Popen(
                                [python_executable, '-m', service_module],
                                stdout=log_f,
                                stderr=subprocess.STDOUT,  # Combine stderr with stdout
                                preexec_fn=os.setpgrp,
                                cwd=os.getcwd(),  # Use current working directory to access config files
                                env=os.environ.copy()  # Inherit environment variables
                            )
                    
                    logger.debug(f"Service subprocess logs available at: {log_file}")
                    print(f"Commander Service started with PID: {cls.pid}")
                    ProcessInfo.save(cls.pid, is_running, ngrok_pid)

                except Exception as e:
                    logger.error(f"Failed to start service subprocess: {e}")
                    raise

            else:
                cleanup_done = False

                def cleanup_cloudflare_on_foreground_exit():
                    """Clean up Cloudflare tunnel when foreground service exits."""
                    nonlocal cleanup_done
                    if cleanup_done:
                        return
                    cleanup_done = True

                    try:
                        # Try to load PID from ProcessInfo first (more reliable)
                        try:
                            process_info = ProcessInfo.load()
                            saved_cloudflare_pid = process_info.cloudflare_pid
                        except (KeyboardInterrupt, SystemExit):
                            raise
                        except Exception as e:
                            logger.debug(f"Could not load process info: {e}")
                            saved_cloudflare_pid = None

                        # Kill Cloudflare tunnel if running
                        cf_pid = saved_cloudflare_pid or cloudflare_pid
                        if cf_pid:
                            print(f"Stopping Cloudflare tunnel (PID: {cf_pid})...")
                            try:
                                if ServiceManager.kill_process_by_pid(cf_pid):
                                    print("Cloudflare tunnel stopped")
                                else:
                                    try:
                                        if ServiceManager.kill_cloudflare_processes():
                                            print("Cloudflare tunnel stopped")
                                    except (KeyboardInterrupt, SystemExit):
                                        raise
                                    except Exception as e:
                                        logger.debug(f"Fallback cloudflare cleanup failed: {e}")
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except Exception as e:
                                logger.debug(f"Primary cloudflare cleanup failed: {e}")
                                try:
                                    if ServiceManager.kill_cloudflare_processes():
                                        print("Cloudflare tunnel stopped")
                                except (KeyboardInterrupt, SystemExit):
                                    raise
                                except Exception as e:
                                    logger.debug(f"Fallback cloudflare cleanup also failed: {e}")
                        else:
                            try:
                                if ServiceManager.kill_cloudflare_processes():
                                    print("Cloudflare tunnel stopped")
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except Exception as e:
                                logger.debug(f"Fallback cloudflare cleanup failed: {e}")

                        # Clear process info
                        try:
                            ProcessInfo.clear()
                        except (KeyboardInterrupt, SystemExit):
                            raise
                        except Exception as e:
                            logger.debug(f"Could not clear process info: {e}")

                    except (KeyboardInterrupt, SystemExit):
                        logger.info("Cloudflare cleanup interrupted by user or system")
                        raise
                    except Exception as e:
                        print(f"Unexpected error during Cloudflare cleanup: {e}")
                        logger.error(f"Unexpected error during Cloudflare cleanup: {e}")

                def foreground_signal_handler(signum, frame):
                    """Handle interrupt signals in foreground mode."""
                    cleanup_cloudflare_on_foreground_exit()
                    sys.exit(0)

                # Set up signal handlers for foreground mode
                import signal
                signal.signal(signal.SIGINT, foreground_signal_handler)   # Ctrl+C
                signal.signal(signal.SIGTERM, foreground_signal_handler)  # Termination

                from ...service.app import create_app
                cls._flask_app = create_app()
                cls._is_running = True

                ProcessInfo.save(os.getpid(), is_running, ngrok_pid)
                ssl_context = ServiceManager.get_ssl_context(config_data)
                
                try:
                    cls._flask_app.run(
                        host='0.0.0.0',
                        port=port,
                        ssl_context=ssl_context
                    )
                finally:
                    cleanup_cloudflare_on_foreground_exit()

            # Save the process ID for future reference
            ProcessInfo.save(cls.pid, is_running, ngrok_pid, cloudflare_pid)
            
        except FileNotFoundError:
            logging.info("Error: Service configuration file not found. Please use 'service-create' command to create a service_config file.")
            return
        except Exception as e:
            logger.error(f"Error: Failed to start Commander Service")
            logger.error(f"Reason: {e}")
            cls._handle_shutdown()

    @classmethod
    def stop_service(cls) -> None:
        """Stop the service if running."""
        process_info = ProcessInfo.load()
        
        logger.debug(f"Loaded process info - Service PID: {process_info.pid}, Ngrok PID: {process_info.ngrok_pid}, Cloudflare PID: {process_info.cloudflare_pid}")
        
        if not process_info.pid:
            print("Error: No running service found to stop")
            return

        try:
            # Stop ngrok process first if it exists
            ngrok_stopped = False
            if process_info.ngrok_pid:
                try:
                    logger.debug(f"Attempting to stop ngrok process (PID: {process_info.ngrok_pid})")
                    
                    # Check if ngrok process is actually running first
                    try:
                        ngrok_process = psutil.Process(process_info.ngrok_pid)
                        logger.debug(f"Ngrok process {process_info.ngrok_pid} is running: {ngrok_process.name()}")
                    except psutil.NoSuchProcess:
                        logger.debug(f"Ngrok process {process_info.ngrok_pid} is not running")
                        
                    logger.debug(f"Calling kill_process_by_pid for ngrok PID {process_info.ngrok_pid}")
                    if ServiceManager.kill_process_by_pid(process_info.ngrok_pid):
                        # Verify that we actually killed an ngrok process, not just any process
                        logger.debug(f"Verifying that PID {process_info.ngrok_pid} was actually ngrok...")
                        try:
                            # Check if there are still any ngrok processes running
                            ngrok_still_running = False
                            for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
                                try:
                                    if proc.info['name'] and 'ngrok' in proc.info['name'].lower():
                                        ngrok_still_running = True
                                        logger.debug(f"Found remaining ngrok process: PID {proc.info['pid']}")
                                        break
                                    elif proc.info['cmdline']:
                                        cmdline_str = ' '.join(proc.info['cmdline'])
                                        if 'ngrok' in cmdline_str.lower() and 'http' in cmdline_str.lower():
                                            ngrok_still_running = True
                                            logger.debug(f"Found remaining ngrok process: PID {proc.info['pid']}")
                                            break
                                except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                                    continue
                            
                            if not ngrok_still_running:
                                logger.debug(f"Ngrok process stopped (PID: {process_info.ngrok_pid})")
                                print("Ngrok tunnel stopped")
                                ngrok_stopped = True
                            else:
                                logger.warning(f"Wrong PID was stored for ngrok process: {process_info.ngrok_pid}")
                                print("Warning: Stored ngrok PID was incorrect, will use fallback cleanup")
                        except Exception as e:
                            logger.debug(f"Error verifying ngrok termination: {e}")
                    else:
                        logger.warning(f"Failed to stop ngrok process (PID: {process_info.ngrok_pid})")
                except Exception as e:
                    logger.warning(f"Error stopping ngrok process: {str(e)}")
            else:
                logger.debug("No ngrok PID found in process info")
            
            # Fallback: Try to kill any remaining ngrok processes
            if not ngrok_stopped:
                logger.debug("Attempting fallback ngrok cleanup...")
                if ServiceManager.kill_ngrok_processes():
                    print("Ngrok tunnel stopped (via cleanup)")
                    ngrok_stopped = True
            
            if not ngrok_stopped:
                logger.debug("No ngrok processes found to stop")

            # Stop Cloudflare tunnel process if it exists
            cloudflare_stopped = False
            if process_info.cloudflare_pid:
                try:
                    logger.debug(f"Attempting to stop Cloudflare tunnel process (PID: {process_info.cloudflare_pid})")

                    # Check if Cloudflare process is actually running first
                    try:
                        cloudflare_process = psutil.Process(process_info.cloudflare_pid)
                        logger.debug(f"Cloudflare tunnel process {process_info.cloudflare_pid} is running: {cloudflare_process.name()}")
                    except psutil.NoSuchProcess:
                        logger.debug(f"Cloudflare tunnel process {process_info.cloudflare_pid} is not running")

                    logger.debug(f"Calling kill_process_by_pid for Cloudflare PID {process_info.cloudflare_pid}")
                    if ServiceManager.kill_process_by_pid(process_info.cloudflare_pid):
                        logger.debug(f"Cloudflare tunnel process stopped (PID: {process_info.cloudflare_pid})")
                        print("Cloudflare tunnel stopped")
                        cloudflare_stopped = True
                    else:
                        logger.warning(f"Failed to stop Cloudflare tunnel process (PID: {process_info.cloudflare_pid})")
                except Exception as e:
                    logger.warning(f"Error stopping Cloudflare tunnel process: {str(e)}")
            else:
                logger.debug("No Cloudflare tunnel PID found in process info")

            # Fallback: Try to kill any remaining cloudflared processes
            if not cloudflare_stopped:
                logger.debug("Attempting fallback Cloudflare tunnel cleanup...")
                if ServiceManager.kill_cloudflare_processes():
                    print("Cloudflare tunnel stopped (via cleanup)")
                    cloudflare_stopped = True

            if not cloudflare_stopped:
                logger.debug("No Cloudflare tunnel processes found to stop")

            # Stop the main service process
            if ServiceManager.kill_process_by_pid(process_info.pid):
                logger.debug(f"Commander Service stopped (PID: {process_info.pid})")
                print("Service stopped successfully")
                process = psutil.Process(process_info.pid)
                process.terminate()

                if process_info.terminal and process_info.terminal != TerminalHandler.get_terminal_info():
                    TerminalHandler.notify_other_terminal(process_info.terminal)

        except (psutil.NoSuchProcess, ProcessLookupError):
            print("Error: No running service found to stop")
        except Exception as e:
            logger.error(f"Error stopping service: {str(e)}")

        cls._handle_shutdown()

    @staticmethod
    def get_status() -> str:
        """Get current service status."""
        process_info = ProcessInfo.load()
        if process_info.pid and process_info.is_running:
            try:
                psutil.Process(process_info.pid)
                terminal = process_info.terminal or "unknown terminal"
                status = f"Commander Service is Running (PID: {process_info.pid}, Terminal: {terminal})"
                
                # Check ngrok status if available
                if process_info.ngrok_pid:
                    try:
                        psutil.Process(process_info.ngrok_pid)
                        # Try to get the current ngrok URL dynamically
                        try:
                            from ..util.tunneling import get_ngrok_url_from_api
                            current_url = get_ngrok_url_from_api(max_retries=1, retry_delay=0.5)
                            if current_url:
                                status += f"\nNgrok tunnel is Running (PID: {process_info.ngrok_pid}, URL: {current_url})"
                            else:
                                status += f"\nNgrok tunnel is Running (PID: {process_info.ngrok_pid})"
                        except Exception:
                            status += f"\nNgrok tunnel is Running (PID: {process_info.ngrok_pid})"
                    except psutil.NoSuchProcess:
                        status += f"\nNgrok tunnel is Stopped (was PID: {process_info.ngrok_pid})"
                
                # Check Cloudflare tunnel status if available
                if process_info.cloudflare_pid:
                    try:
                        psutil.Process(process_info.cloudflare_pid)
                        status += f"\nCloudflare tunnel is Running (PID: {process_info.cloudflare_pid})"
                    except psutil.NoSuchProcess:
                        status += f"\nCloudflare tunnel is Stopped (was PID: {process_info.cloudflare_pid})"

                logger.debug(f"Service status check: {status}")
                return status
            except psutil.NoSuchProcess:
                ProcessInfo.clear()
                pass
        else:
            status = "No Commander Service is running currently"
            return status
        status = "Commander Service is Stopped"
        logger.debug(f"Service status check: {status}")
        return status
    
    @staticmethod
    def kill_process_by_pid(pid: int):
        """Terminates a process by PID without using psutil."""
        logger.debug(f"Attempting to kill process {pid}")
        try:
            if sys.platform.startswith("win"):  #  Windows
                logger.debug(f"Using Windows taskkill for PID {pid}")
                subprocess.run(["taskkill", "/PID", str(pid), "/F"], check=True)
                return True
            else:  #  Linux & macOS
                try:
                    logger.debug(f"Sending SIGTERM to PID {pid}")
                    os.kill(pid, signal.SIGTERM)
                    # Wait a moment for graceful termination
                    import time
                    time.sleep(0.5)
                    # Check if process is still running
                    try:
                        logger.debug(f"Checking if PID {pid} is still running")
                        os.kill(pid, 0)  # This doesn't kill, just checks if process exists
                        # If we get here, process is still running, force kill it
                        logger.debug(f"Process {pid} still running, sending SIGKILL")
                        os.kill(pid, signal.SIGKILL)
                        logger.debug(f"Sent SIGKILL to PID {pid}")
                    except ProcessLookupError:
                        # Process has terminated gracefully
                        logger.debug(f"Process {pid} terminated gracefully after SIGTERM")
                        pass
                    return True
                except ProcessLookupError:
                    # Process doesn't exist
                    logger.debug(f"Process {pid} doesn't exist")
                    return True
        except ProcessLookupError:
            logger.warning(f"No process found with PID {pid}. It may have already exited.")
            return True
        except Exception as e:
            logger.error(f"Error terminating process {pid}: {str(e)}")
        return False

    @staticmethod
    def kill_ngrok_processes():
        """Kill all ngrok processes as a fallback method."""
        killed_count = 0
        try:
            logger.debug("Looking for ngrok processes to kill...")
            
            # Find all ngrok processes
            for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
                try:
                    if proc.info['name'] and 'ngrok' in proc.info['name'].lower():
                        pid = proc.info['pid']
                        logger.debug(f"Found ngrok process by name: PID {pid}")
                        if ServiceManager.kill_process_by_pid(pid):
                            killed_count += 1
                    elif proc.info['cmdline']:
                        # Check if ngrok is in the command line
                        cmdline_str = ' '.join(proc.info['cmdline'])
                        if 'ngrok' in cmdline_str.lower() and 'http' in cmdline_str.lower():
                            pid = proc.info['pid']
                            logger.debug(f"Found ngrok process by cmdline: PID {pid}")
                            if ServiceManager.kill_process_by_pid(pid):
                                killed_count += 1
                except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                    continue
            
            if killed_count > 0:
                logger.info(f"Killed {killed_count} ngrok processes")
                return True
            else:
                logger.debug("No ngrok processes found to kill")
                return False
                
        except Exception as e:
            logger.error(f"Exception while killing ngrok processes: {str(e)}")
            return False

    @staticmethod
    def kill_cloudflare_processes():
        """Kill all cloudflared processes as a fallback method."""
        killed_count = 0
        try:
            logger.debug("Looking for cloudflared processes to kill...")

            # Find all cloudflared processes
            for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
                try:
                    if proc.info['name'] and 'cloudflared' in proc.info['name'].lower():
                        pid = proc.info['pid']
                        logger.debug(f"Found cloudflared process by name: PID {pid}")
                        if ServiceManager.kill_process_by_pid(pid):
                            killed_count += 1
                    elif proc.info['cmdline']:
                        # Check if cloudflared is in the command line
                        cmdline_str = ' '.join(proc.info['cmdline'])
                        if 'cloudflared' in cmdline_str.lower() and ('tunnel' in cmdline_str.lower() or 'http' in cmdline_str.lower()):
                            pid = proc.info['pid']
                            logger.debug(f"Found cloudflared process by cmdline: PID {pid}")
                            if ServiceManager.kill_process_by_pid(pid):
                                killed_count += 1
                except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                    continue

            if killed_count > 0:
                logger.info(f"Killed {killed_count} cloudflared processes")
                return True
            else:
                logger.debug("No cloudflared processes found to kill")
                return False

        except Exception as e:
            logger.error(f"Exception while killing cloudflared processes: {str(e)}")
            return False



from __future__ import annotations
import argparse
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ...display import bcolors
from ..pam.router_helper import router_get_connected_gateways
from ...discovery_common.rule import Rules
from ...discovery_common.types import RuleTypeEnum
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from ...discovery_common.types import RuleItem


class PAMGatewayActionDiscoverRuleListCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-discover-rule-list')
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name of UID.')
    parser.add_argument('--search', '-s', required=False, dest='search', action='store',
                        help='Search for rules.')

    def get_parser(self):
        return PAMGatewayActionDiscoverRuleListCommand.parser

    @staticmethod
    def print_rule_table(rule_list: List[RuleItem]):

        print("")
        print(f"{bcolors.HEADER}{'Rule ID'.ljust(15, ' ')} "
              f"{'Action'.ljust(6, ' ')} "
              f"{'Priority'.ljust(8, ' ')} "
              f"{'Case'.ljust(12, ' ')} "
              f"{'Added'.ljust(19, ' ')} "
              f"{'Shared Folder UID'.ljust(22, ' ')} "
              "Rule"
              f"{bcolors.ENDC}")

        print(f"{''.ljust(15, '=')} "
              f"{''.ljust(6, '=')} "
              f"{''.ljust(8, '=')} "
              f"{''.ljust(12, '=')} "
              f"{''.ljust(19, '=')} "
              f"{''.ljust(22, '=')} "
              f"{''.ljust(10, '=')} ")

        for rule in rule_list:
            if rule.case_sensitive is True:
                ignore_case_str = "Sensitive"
            else:
                ignore_case_str = "Insensitive"

            shared_folder_uid = ""
            if rule.shared_folder_uid is not None:
                shared_folder_uid = rule.shared_folder_uid
            print(f"{bcolors.OKGREEN}{rule.rule_id.ljust(14, ' ')}{bcolors.ENDC} "
                  f"{rule.action.value.ljust(6, ' ')} "
                  f"{str(rule.priority).rjust(8, ' ')} "
                  f"{ignore_case_str.ljust(12, ' ')} "
                  f"{rule.added_ts_str.ljust(19, ' ')} "
                  f"{shared_folder_uid.ljust(22, ' ')} "
                  f"{Rules.make_action_rule_statement_str(rule.statement)}")

    def execute(self, params, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        gateway = kwargs.get("gateway")
        gateway_context = GatewayContext.from_gateway(params, gateway)
        if gateway_context is None:
            print(f'{bcolors.FAIL}Discovery job gateway [{gateway}] was not found.{bcolors.ENDC}')
            return

        rules = Rules(record=gateway_context.configuration, params=params)
        rule_list = rules.rule_list(rule_type=RuleTypeEnum.ACTION,
                                    search=kwargs.get("search"))  # type: List[RuleItem]
        if len(rule_list) == 0:
            print(f"{bcolors.FAIL}There are no rules. Use 'pam action discovery rule add' "
                  f"to create rules.{bcolors.ENDC}")
            return

        self.print_rule_table(rule_list=rule_list)



import argparse
import logging
import sys

from keepercommander import api, vault_extensions, vault
from keepercommander.__main__ import get_params_from_config
from keepercommander.commands import convert

parser = argparse.ArgumentParser(description='Converts General records with attachments to Typed records with fileRef fields')
parser.add_argument('--debug', action='store_true', help='Enables debug logging')
opts, flags = parser.parse_known_args(sys.argv[1:])

logging.basicConfig(level=logging.DEBUG if opts.debug is True else logging.WARNING, format='%(message)s')

my_params = get_params_from_config('')

api.login(my_params)
if not my_params.session_token:
    exit(1)

api.sync_down(my_params)

record_uids = []

for record in vault_extensions.find_records(my_params, record_version=2):
    if isinstance(record, vault.PasswordRecord):
        if isinstance(record.attachments, list) and len(record.attachments) > 0:
            record_uids.append(record.record_uid)
            if len(record_uids) >= 200:
                break

if len(record_uids) > 0:
    if len(record_uids) >= 200:
        print('Found more than 200 records with attachments. Only the first 200 will be converted. Repeat this command to convert the rest.')
    else:
        print(f'Found {len(record_uids)} record(s) with attachments. Converting...')
    cmd = convert.ConvertCommand()
    kwargs = {
        'record-uid-name-patterns': record_uids
    }
    cmd.execute(my_params, **kwargs)
else:
    print('No records found')



from __future__ import annotations
import argparse
from ..discover import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ... import vault
from . import get_plugins_map
from ...discovery_common.record_link import RecordLink
from ...discovery_common.constants import PAM_USER, PAM_MACHINE, PAM_DATABASE, PAM_DIRECTORY
from ...discovery_common.types import UserAclRotationSettings
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ...vault import TypedRecord
    from ...params import KeeperParams


class PAMActionSaasAddCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam action saas add')

    parser.add_argument('--user-uid', '-u', required=True, dest='user_uid', action='store',
                        help='The UID of the User record')
    parser.add_argument('--config-record-uid', '-c', required=True, dest='config_record_uid',
                        action='store', help='The UID of the record that has SaaS configuration')
    parser.add_argument('--resource-uid', '-r', required=False, dest='resource_uid', action='store',
                        help='The UID of the Resource record, if needed.')

    def get_parser(self):
        return PAMActionSaasAddCommand.parser

    def execute(self, params: KeeperParams, **kwargs):

        user_uid = kwargs.get("user_uid")  # type: str
        resource_uid = kwargs.get("resource_uid")  # type: str
        config_record_uid = kwargs.get("config_record_uid")   # type: str

        print("")

        # Check to see if the record exists.
        user_record = vault.KeeperRecord.load(params, user_uid)  # type: Optional[TypedRecord]
        if user_record is None:
            print(self._f("The user record does not exists."))
            return

        # Make sure this user is a PAM User.
        if user_record.record_type != PAM_USER:
            print(self._f("The user record is not a PAM User."))
            return

        record_rotation = params.record_rotation_cache.get(user_record.record_uid)
        if record_rotation is not None:
            configuration_uid = record_rotation.get("configuration_uid")
        else:
            print(self._f("The user record does not have any rotation settings."))
            return

        if configuration_uid is None:
            print(self._f("The user record does not have the configuration record set in the rotation settings."))
            return

        gateway_context = GatewayContext.from_configuration_uid(params, configuration_uid)

        if gateway_context is None:
            print(self._f("The user record does not have the set gateway"))
            return

        plugins = get_plugins_map(params, gateway_context)
        if plugins is None:
            return

        # Check to see if the config record exists.
        config_record = vault.KeeperRecord.load(params, config_record_uid)  # type: Optional[TypedRecord]
        if config_record is None:
            print(self._f("The SaaS configuration record does not exists."))
            return

        # Make sure this config is a Login record.

        if config_record.record_type != "login":
            print(self._f("The SaaS configuration record is not a Login record."))
            return

        plugin_name_field = next((x for x in config_record.custom if x.label == "SaaS Type"), None)
        if plugin_name_field is None:
            print(self._f("The SaaS configuration record is missing the custom field label 'SaaS Type'"))
            return

        plugin_name = None
        if plugin_name_field.value is not None and len(plugin_name_field.value) > 0:
            plugin_name = plugin_name_field.value[0]

        if plugin_name is None:
            print(self._f("The SaaS configuration record's custom field label 'SaaS Type' does not have a value."))
            return

        if plugin_name not in plugins:
            print(self._f("The SaaS configuration record's custom field label 'SaaS Type' is not supported by the "
                          "gateway or the value is not correct."))
            return

        plugin = plugins[plugin_name]

        # Make sure the SaaS configuration record has correct custom fields.
        missing_fields = []
        for field in plugin.fields:
            if field.required is True and field.default_value is None:
                found = next((x for x in config_record.custom if x.label == field.label), None)
                if not found:
                    missing_fields.append(field.label.strip())

        if len(missing_fields) > 0:
            print(self._f("The SaaS configuration record is missing the following required custom fields: "
                          f'{", ".join(missing_fields)}'))
            return

        parent_uid = gateway_context.configuration_uid

        # Not sure if SaaS type rotation should be limited to NOOP rotation.
        # Allow a resource record to be used.
        if resource_uid is not None:
            # Check to see if the record exists.
            resource_record = vault.KeeperRecord.load(params, resource_uid)  # type: Optional[TypedRecord]
            if resource_record is None:
                print(self._f("The resource record does not exists."))
                return

            # Make sure this user is a PAM User.
            if user_record.record_type in [PAM_MACHINE, PAM_DATABASE, PAM_DIRECTORY]:
                print(self._f("The resource record does not have the correct record type."))
                return

            parent_uid = resource_uid

        record_link = RecordLink(record=gateway_context.configuration, params=params, fail_on_corrupt=False)
        acl = record_link.get_acl(user_uid, parent_uid)
        if acl is None:
            if resource_uid is not None:
                print(self._f("There is no relationship between the user and the resource record."))
            else:
                print(self._f("There is no relationship between the user and the configuration record."))
            return

        if acl.rotation_settings is None:
            acl.rotation_settings = UserAclRotationSettings()

        if resource_uid is not None and acl.rotation_settings.noop is True:
            print(self._f("The rotation is flagged as No Operation, however you passed in a resource record. "
                          "This combination is not allowed."))
            return

        # If there is a resource record, it not NOOP.
        # If there is NO resource record, it is NOOP.
        # However, if this is an IAM User, don't set the NOOP
        if acl.is_iam_user is False:
            acl.rotation_settings.noop = resource_uid is None

        # PyCharm didn't like appending directly, so do this stupid thing.
        record_uid_list = acl.rotation_settings.saas_record_uid_list

        # Make sure we are not re-adding the same SaaS config.
        if config_record_uid in record_uid_list:
            print(self._f("The SaaS configuration record is already being used for this user."))
            return

        record_uid_list.append(config_record_uid)
        acl.rotation_settings.saas_record_uid_list = record_uid_list

        record_link.belongs_to(user_uid, parent_uid, acl=acl)
        record_link.save()

        print(self._gr(f"Added {plugin_name} rotation to the user record."))
        print("")



from typing import Optional

from ... import vault, record_facades, record_types


class PamConfigurationRecordFacade(record_facades.TypedRecordFacade):
    _controller_uid_getter = record_facades.string_element_getter('_pam_resources', 'controllerUid')
    _controller_uid_setter = record_facades.string_element_setter('_pam_resources', 'controllerUid')
    _folder_uid_getter = record_facades.string_element_getter('_pam_resources', 'folderUid')
    _folder_uid_setter = record_facades.string_element_setter('_pam_resources', 'folderUid')
    _resource_ref_getter = record_facades.list_element_getter('_pam_resources', 'resourceRef')
    _file_ref_getter = record_facades.string_list_getter('_file_ref')

    def __init__(self):
        super(PamConfigurationRecordFacade, self).__init__()
        self._pam_resources = None    # type: Optional[vault.TypedField]
        self._port_mapping = None     # type: Optional[vault.TypedField]
        self._file_ref = None         # type: Optional[vault.TypedField]

    def load_typed_fields(self):
        if self.record:
            self._pam_resources = next((x for x in self.record.fields if x.type == 'pamResources'), None)
            if not self._pam_resources:
                self._pam_resources = vault.TypedField.new_field('pamResources', [])
                self.record.fields.append(self._pam_resources)

            if len(self._pam_resources.value) > 0:
                if not isinstance(self._pam_resources.value[0], dict):
                    self._pam_resources.value.clear()

            if len(self._pam_resources.value) == 0:
                if 'pamResources' in record_types.FieldTypes and isinstance(record_types.FieldTypes['pamResources'].value, dict):
                    value = record_types.FieldTypes['pamResources'].value.copy()
                else:
                    value = {}
                self._pam_resources.value.append(value)

            self._port_mapping = next((x for x in self.record.fields
                                       if x.type == 'multiline' and x.label == 'portMapping'), None)
            if self._port_mapping is None:
                self._port_mapping = vault.TypedField.new_field('multiline', [], field_label='portMapping')
                self.record.fields.append(self._port_mapping)

            self._file_ref = next((x for x in self.record.fields if x.type == 'fileRef' and x.label == 'rotationScripts'), None)
            if self._file_ref is None:
                self._file_ref = vault.TypedField.new_field('fileRef', [], field_label='rotationScripts')
                self.record.fields.append(self._file_ref)
        else:
            self._pam_resources = None
            self._port_mapping = None
            self._file_ref = None

        super(PamConfigurationRecordFacade, self).load_typed_fields()

    @property
    def controller_uid(self):
        return PamConfigurationRecordFacade._controller_uid_getter(self)

    @controller_uid.setter
    def controller_uid(self, value):
        PamConfigurationRecordFacade._controller_uid_setter(self, value)

    @property
    def folder_uid(self):
        return PamConfigurationRecordFacade._folder_uid_getter(self)

    @folder_uid.setter
    def folder_uid(self, value):
        PamConfigurationRecordFacade._folder_uid_setter(self, value)

    @property
    def resource_ref(self):
        return PamConfigurationRecordFacade._resource_ref_getter(self)

    @property
    def rotation_scripts(self):
        return PamConfigurationRecordFacade._file_ref_getter(self)



from ....commands.tunnel.port_forward.tunnel_helpers import generate_random_bytes, get_config_uid
from ....keeper_dag import DAG, EdgeType
from ....keeper_dag.connection.commander import Connection
from ....keeper_dag.types import RefType
from ....keeper_dag.vertex import DAGVertex
from ....display import bcolors
from ....vault import PasswordRecord


def get_vertex_content(vertex):
    return_content = None
    if vertex is None:
        return return_content
    try:
        return_content = vertex.content_as_dict
    except Exception as e:
        import logging
        logging.debug(f"Error getting vertex content: {e}")
        return_content = None
    return return_content


class TunnelDAG:
    def __init__(self, params, encrypted_session_token, encrypted_transmission_key, record_uid: str, is_config=False):
        config_uid = None
        if not is_config:
            config_uid = get_config_uid(params, encrypted_session_token, encrypted_transmission_key, record_uid)
        if not config_uid:
            config_uid = record_uid
        self.record = PasswordRecord()
        self.record.record_uid = config_uid
        self.record.record_key = generate_random_bytes(32)
        self.encrypted_session_token = encrypted_session_token
        self.encrypted_transmission_key = encrypted_transmission_key
        self.conn = Connection(params=params, encrypted_transmission_key=self.encrypted_transmission_key,
                               encrypted_session_token=self.encrypted_session_token
                               )
        self.linking_dag = DAG(conn=self.conn, record=self.record, graph_id=0)
        try:
            self.linking_dag.load()
        except Exception as e:
            import logging
            logging.debug(f"Error loading config: {e}")

    def resource_belongs_to_config(self, resource_uid):
        if not self.linking_dag.has_graph:
            return False
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        return resource_vertex and config_vertex.has(resource_vertex, EdgeType.LINK)

    def user_belongs_to_config(self, user_uid):
        if not self.linking_dag.has_graph:
            return False
        user_vertex = self.linking_dag.get_vertex(user_uid)
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        res_content = False
        if user_vertex and config_vertex and config_vertex.has(user_vertex, EdgeType.ACL):
            acl_edge = user_vertex.get_edge(config_vertex, EdgeType.ACL)
            _content = acl_edge.content_as_dict
            res_content = _content.get('belongs_to', False) if _content else False
        return res_content

    def check_tunneling_enabled_config(self, enable_connections=None, enable_tunneling=None,
                                       enable_rotation=None, enable_session_recording=None,
                                       enable_typescript_recording=None, remote_browser_isolation=None):
        if not self.linking_dag.has_graph:
            return False
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        content = get_vertex_content(config_vertex)
        if content is None or not content.get('allowedSettings'):
            return False

        allowed_settings = content['allowedSettings']
        if enable_connections and not allowed_settings.get("connections"):
            return False
        if enable_tunneling and not allowed_settings.get("portForwards"):
            return False
        if enable_rotation and not allowed_settings.get("rotation"):
            return False
        if allowed_settings.get("connections") and allowed_settings["connections"]:
            if enable_session_recording and not allowed_settings.get("sessionRecording"):
                return False
            if enable_typescript_recording and not allowed_settings.get("typescriptRecording"):
                return False
        if remote_browser_isolation and not allowed_settings.get("remoteBrowserIsolation"):
            return False
        return True

    @staticmethod
    def _convert_allowed_setting(value):
        """Converts on/off/default|any to True/False/None"""
        if value is None or isinstance(value, bool):
            return value
        return {"on": True, "off": False}.get(str(value).lower(), None)

    def edit_tunneling_config(self, connections=None, tunneling=None,
                              rotation=None, session_recording=None,
                              typescript_recording=None,
                              remote_browser_isolation=None):
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        if config_vertex is None:
            config_vertex = self.linking_dag.add_vertex(uid=self.record.record_uid, vertex_type=RefType.PAM_NETWORK)

        if config_vertex.vertex_type != RefType.PAM_NETWORK:
            config_vertex.vertex_type = RefType.PAM_NETWORK
        content = get_vertex_content(config_vertex)
        if content and content.get('allowedSettings'):
            allowed_settings = dict(content['allowedSettings'])
            del content['allowedSettings']
            content = {'allowedSettings': allowed_settings}

        if content is None:
            content = {'allowedSettings': {}}
        if 'allowedSettings' not in content:
            content['allowedSettings'] = {}

        allowed_settings = content['allowedSettings']
        dirty = False

        # When no value in allowedSettings: client will substitute with default
        # rotation defaults to True, everything else defaults to False

        # switching to 3-state on/off/default: on/true, off/false,
        # None = Keep existing, 'default' = Reset to default (remove from dict)
        if connections is not None:
            connections = self._convert_allowed_setting(connections)
            if connections != allowed_settings.get("connections", None):
                dirty = True
                if connections is None:
                    allowed_settings.pop("connections", None)
                else:
                    allowed_settings["connections"] = connections

        if tunneling is not None:
            tunneling = self._convert_allowed_setting(tunneling)
            if tunneling != allowed_settings.get("portForwards", None):
                dirty = True
                if tunneling is None:
                    allowed_settings.pop("portForwards", None)
                else:
                    allowed_settings["portForwards"] = tunneling

        if rotation is not None:
            rotation = self._convert_allowed_setting(rotation)
            if rotation != allowed_settings.get("rotation", None):
                dirty = True
                if rotation is None:
                    allowed_settings.pop("rotation", None)
                else:
                    allowed_settings["rotation"] = rotation

        if session_recording is not None:
            session_recording = self._convert_allowed_setting(session_recording)
            if session_recording != allowed_settings.get("sessionRecording", None):
                dirty = True
                if session_recording is None:
                    allowed_settings.pop("sessionRecording", None)
                else:
                    allowed_settings["sessionRecording"] = session_recording

        if typescript_recording is not None:
            typescript_recording = self._convert_allowed_setting(typescript_recording)
            if typescript_recording != allowed_settings.get("typescriptRecording", None):
                dirty = True
                if typescript_recording is None:
                    allowed_settings.pop("typescriptRecording", None)
                else:
                    allowed_settings["typescriptRecording"] = typescript_recording

        if remote_browser_isolation is not None:
            remote_browser_isolation = self._convert_allowed_setting(remote_browser_isolation)
            if remote_browser_isolation != allowed_settings.get("remoteBrowserIsolation", None):
                dirty = True
                if remote_browser_isolation is None:
                    allowed_settings.pop("remoteBrowserIsolation", None)
                else:
                    allowed_settings["remoteBrowserIsolation"] = remote_browser_isolation

        if dirty:
            config_vertex.add_data(content=content, path='meta', needs_encryption=False)
            self.linking_dag.save()

    def get_all_owners(self, uid):
        owners = []
        if self.linking_dag.has_graph:
            vertex = self.linking_dag.get_vertex(uid)
            if vertex:
                owners = [owner.uid for owner in vertex.belongs_to_vertices()]
        return owners

    def user_belongs_to_resource(self, user_uid, resource_uid):
        user_vertex = self.linking_dag.get_vertex(user_uid)
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        res_content = False
        if user_vertex and resource_vertex and resource_vertex.has(user_vertex, EdgeType.ACL):
            acl_edge = user_vertex.get_edge(resource_vertex, EdgeType.ACL)
            _content = acl_edge.content_as_dict
            res_content = _content.get('belongs_to', False) if _content else False
        return res_content

    def get_resource_uid(self, user_uid):
        if not self.linking_dag.has_graph:
            return None
        resources = self.get_all_owners(user_uid)
        if len(resources) > 0:
            for resource in resources:
                if self.user_belongs_to_resource(user_uid, resource):
                    return resource
        return None

    def link_resource_to_config(self, resource_uid):
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        if config_vertex is None:
            config_vertex = self.linking_dag.add_vertex(uid=self.record.record_uid)

        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        if resource_vertex is None:
            resource_vertex = self.linking_dag.add_vertex(uid=resource_uid)

        if not config_vertex.has(resource_vertex, EdgeType.LINK):
            resource_vertex.belongs_to(config_vertex, EdgeType.LINK)
            self.linking_dag.save()

    def link_user_to_config(self, user_uid):
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        if config_vertex is None:
            config_vertex = self.linking_dag.add_vertex(uid=self.record.record_uid)
        self.link_user(user_uid, config_vertex, belongs_to=True, is_iam_user=True)

    def link_user_to_config_with_options(self, user_uid, is_admin=None, belongs_to=None, is_iam_user=None):
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        if config_vertex is None:
            config_vertex = self.linking_dag.add_vertex(uid=self.record.record_uid)

        # self.link_user(user_uid, config_vertex, is_admin, belongs_to, is_iam_user)
        source_vertex = config_vertex
        user_vertex = self.linking_dag.get_vertex(user_uid)
        if user_vertex is None:
            user_vertex = self.linking_dag.add_vertex(uid=user_uid, vertex_type=RefType.PAM_USER)

        # switching to 3-state on/off/default: on/true, off/false,
        # None = Keep existing, 'default' = Reset to default (remove from dict)
        states = {'on': True, 'off': False, 'default': '', 'none': None}

        content = {
            "belongs_to": states.get(str(belongs_to).lower()),
            "is_admin": states.get(str(is_admin).lower()),
            "is_iam_user": states.get(str(is_iam_user).lower())
        }
        if user_vertex.vertex_type != RefType.PAM_USER:
            user_vertex.vertex_type = RefType.PAM_USER

        dirty = False
        if source_vertex.has(user_vertex, EdgeType.ACL):
            acl_edge = user_vertex.get_edge(source_vertex, EdgeType.ACL)
            existing_content = acl_edge.content_as_dict or {}
            old_content = existing_content.copy()
            for key in list(existing_content.keys()):
                if content.get(key) is not None:
                    if content[key] == '':
                        existing_content.pop(key)
                    elif content[key] in (True, False):
                        existing_content[key] = content[key]
            content = {k: v for k, v in content.items() if v not in (None, '')}
            for k, v in content.items():
                existing_content.setdefault(k, v)
            if existing_content != old_content:
                dirty = True

            if dirty:
                user_vertex.belongs_to(source_vertex, EdgeType.ACL, content=existing_content)
                # user_vertex.add_data(content=existing_content, needs_encryption=False)
                self.linking_dag.save()
        else:
            content = {k: v for k, v in content.items() if v not in (None, '')}
            user_vertex.belongs_to(source_vertex, EdgeType.ACL, content=content)
            self.linking_dag.save()

    def unlink_user_from_resource(self, user_uid, resource_uid) -> bool:
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        if resource_vertex is None or not self.resource_belongs_to_config(resource_uid):
            print(f"{bcolors.FAIL}Resource {resource_uid} does not belong to the configuration{bcolors.ENDC}")
            return False

        user_vertex = self.linking_dag.get_vertex(user_uid)
        if user_vertex is None or user_vertex.vertex_type != RefType.PAM_USER:
            return False

        if resource_vertex.has(user_vertex, EdgeType.ACL):
            acl_edge = user_vertex.get_edge(resource_vertex, EdgeType.ACL)
            edge_content = acl_edge.content_as_dict or {}
            link_keys = ('belongs_to', 'is_admin')  # "is_iam_user"
            dirty = any(key in link_keys for key in edge_content)
            if dirty:
                for link_key in link_keys:
                    edge_content.pop(link_key, None)
                user_vertex.belongs_to(resource_vertex, EdgeType.ACL, content=edge_content)
                self.linking_dag.save()
                return True

        return False

    def link_user_to_resource(self, user_uid, resource_uid, is_admin=None, belongs_to=None):
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        if resource_vertex is None or not self.resource_belongs_to_config(resource_uid):
            print(f"{bcolors.FAIL}Resource {resource_uid} does not belong to the configuration{bcolors.ENDC}")
            return False
        self.link_user(user_uid, resource_vertex, is_admin, belongs_to)
        return None

    def link_user(self, user_uid, source_vertex: DAGVertex, is_admin=None, belongs_to=None, is_iam_user=None):

        user_vertex = self.linking_dag.get_vertex(user_uid)
        if user_vertex is None:
            user_vertex = self.linking_dag.add_vertex(uid=user_uid, vertex_type=RefType.PAM_USER)

        content = {}
        dirty = False
        if belongs_to is not None:
            content["belongs_to"] = bool(belongs_to)
        if is_admin is not None:
            content["is_admin"] = bool(is_admin)
        if is_iam_user is not None:
            content["is_iam_user"] = bool(is_iam_user)

        if user_vertex.vertex_type != RefType.PAM_USER:
            user_vertex.vertex_type = RefType.PAM_USER

        if source_vertex.has(user_vertex, EdgeType.ACL):
            acl_edge = user_vertex.get_edge(source_vertex, EdgeType.ACL)
            existing_content = acl_edge.content_as_dict or {}
            for key in existing_content:
                if key not in content:
                    content[key] = existing_content[key]
            if content != existing_content:
                dirty = True

            if dirty:
                user_vertex.belongs_to(source_vertex, EdgeType.ACL, content=content)
                # user_vertex.add_data(content=content, needs_encryption=False)
                self.linking_dag.save()
        else:
            user_vertex.belongs_to(source_vertex, EdgeType.ACL, content=content)
            self.linking_dag.save()

    def get_all_admins(self):
        if not self.linking_dag.has_graph:
            return []
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        if config_vertex is None:
            return []
        admins = []
        for user_vertex in config_vertex.has_vertices(EdgeType.ACL):
            acl_edge = user_vertex.get_edge(config_vertex, EdgeType.ACL)
            if acl_edge:
                content = acl_edge.content_as_dict
                if content.get('is_admin'):
                    admins.append(user_vertex.uid)
        return admins

    def check_if_resource_has_admin(self, resource_uid):
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        if resource_vertex is None:
            return False
        for user_vertex in resource_vertex.has_vertices(EdgeType.ACL):
            acl_edge = user_vertex.get_edge(resource_vertex, EdgeType.ACL)
            if acl_edge:
                content = acl_edge.content_as_dict
                if content.get('is_admin'):
                    return user_vertex.uid
        return False

    def check_if_resource_allowed(self, resource_uid, setting):
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        content = get_vertex_content(resource_vertex)
        return content.get('allowedSettings', {}).get(setting, False) if content else False

    def get_resource_setting(self, resource_uid: str, settings_name: str, setting: str) -> str:
        # Settings are tri-state (on|off|default) mapped to true|false|missing in JSON
        # When set to "default" (missing from JSON) that means look higher up the hierarchy
        # ex. rotation: user -> machine -> pam_config -> Gobal Default settings
        # Note: Different clients (even different client versions)
        # may have different view on these defaults (Commander, Web Vault, etc.)
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        content = get_vertex_content(resource_vertex)
        res = ''
        if content and isinstance(content, dict):
            if settings_name in content and isinstance(content[settings_name], dict):
                if setting in content[settings_name]:
                    value = content[settings_name][setting]
                    if isinstance(value, bool):
                        res = {True: 'on', False: 'off'}[value]
                    else:
                        res = str(value)
                else:
                    res = 'default'

        return res

    def set_resource_allowed(self, resource_uid, tunneling=None, connections=None, rotation=None,
                             session_recording=None, typescript_recording=None, remote_browser_isolation=None,
                             allowed_settings_name='allowedSettings', is_config=False,
                             v_type: RefType=str(RefType.PAM_MACHINE)):
        v_type = RefType(v_type)
        allowed_ref_types = [RefType.PAM_MACHINE, RefType.PAM_DATABASE, RefType.PAM_DIRECTORY, RefType.PAM_BROWSER]
        if v_type not in allowed_ref_types:
            # default to machine
            v_type = RefType.PAM_MACHINE

        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        if resource_vertex is None:
            resource_vertex = self.linking_dag.add_vertex(uid=resource_uid, vertex_type=v_type)

        if resource_vertex.vertex_type not in allowed_ref_types:
            resource_vertex.vertex_type = v_type
        if is_config:
            resource_vertex.vertex_type = RefType.PAM_NETWORK
        dirty = False
        content = get_vertex_content(resource_vertex)
        if content is None:
            content = {allowed_settings_name: {}}
            dirty = True
        if allowed_settings_name not in content:
            content[allowed_settings_name] = {}
            dirty = True

        settings = content[allowed_settings_name]

        # When no value in allowedSettings: client will substitute with default
        # rotation defaults to True, everything else defaults to False

        # switching to 3-state on/off/default: on/true, off/false,
        # None = Keep existing, 'default' = Reset to default (remove from dict)
        if connections is not None:
            connections = self._convert_allowed_setting(connections)
            if connections != settings.get("connections", None):
                dirty = True
                if connections is None:
                    settings.pop("connections", None)
                else:
                    settings["connections"] = connections

        if tunneling is not None:
            tunneling = self._convert_allowed_setting(tunneling)
            if tunneling != settings.get("portForwards", None):
                dirty = True
                if tunneling is None:
                    settings.pop("portForwards", None)
                else:
                    settings["portForwards"] = tunneling

        if rotation is not None:
            rotation = self._convert_allowed_setting(rotation)
            if rotation != settings.get("rotation", None):
                dirty = True
                if rotation is None:
                    settings.pop("rotation", None)
                else:
                    settings["rotation"] = rotation

        if session_recording is not None:
            session_recording = self._convert_allowed_setting(session_recording)
            if session_recording != settings.get("sessionRecording", None):
                dirty = True
                if session_recording is None:
                    settings.pop("sessionRecording", None)
                else:
                    settings["sessionRecording"] = session_recording

        if typescript_recording is not None:
            typescript_recording = self._convert_allowed_setting(typescript_recording)
            if typescript_recording != settings.get("typescriptRecording", None):
                dirty = True
                if typescript_recording is None:
                    settings.pop("typescriptRecording", None)
                else:
                    settings["typescriptRecording"] = typescript_recording

        if remote_browser_isolation is not None:
            remote_browser_isolation = self._convert_allowed_setting(remote_browser_isolation)
            if remote_browser_isolation != settings.get("remoteBrowserIsolation", None):
                dirty = True
                if remote_browser_isolation is None:
                    settings.pop("remoteBrowserIsolation", None)
                else:
                    settings["remoteBrowserIsolation"] = remote_browser_isolation

        if dirty:
            resource_vertex.add_data(content=content, path='meta', needs_encryption=False)
            self.linking_dag.save()

    def is_tunneling_config_set_up(self, resource_uid):
        if not self.linking_dag.has_graph:
            return False
        resource_vertex = self.linking_dag.get_vertex(resource_uid)
        config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
        return resource_vertex and config_vertex and config_vertex in resource_vertex.belongs_to_vertices()

    def remove_from_dag(self, uid):
        if not self.linking_dag.has_graph:
            return True

        vertex = self.linking_dag.get_vertex(uid)
        if vertex is None:
            return True

        vertex.delete()
        self.linking_dag.save(confirm=True)
        return None

    def print_tunneling_config(self, record_uid, pam_settings=None, config_uid=None):
        if not pam_settings and not config_uid:
            return
        self.linking_dag.load()
        vertex = self.linking_dag.get_vertex(record_uid)
        content = get_vertex_content(vertex)
        config_id = config_uid if config_uid else pam_settings.value[0].get('configUid') if pam_settings else None
        if content and content.get('allowedSettings'):
            allowed_settings = content['allowedSettings']
            print(f"{bcolors.OKGREEN}Settings configured for {record_uid}{bcolors.ENDC}")
            port_forwarding = f"{bcolors.OKBLUE}Enabled" if allowed_settings.get('portForwards') else \
                f"{bcolors.WARNING}Disabled"
            rotation = f"{bcolors.WARNING}Disabled" if (allowed_settings.get('rotation') and not allowed_settings['rotation']) else f"{bcolors.OKBLUE}Enabled"
            print(f"{bcolors.OKGREEN}\tRotation: {rotation}{bcolors.ENDC}")
            print(f"{bcolors.OKGREEN}\tTunneling: {port_forwarding}{bcolors.ENDC}")

            print(f"{bcolors.OKGREEN}Configuration: {config_id} {bcolors.ENDC}")
            if config_id is not None:
                config_vertex = self.linking_dag.get_vertex(self.record.record_uid)
                config_content = get_vertex_content(config_vertex)
                if config_content and config_content.get('allowedSettings'):
                    config_allowed_settings = config_content['allowedSettings']
                    config_port_forwarding = f"{bcolors.OKBLUE}Enabled" if (
                        config_allowed_settings.get('portForwards')) else \
                        f"{bcolors.WARNING}Disabled"
                    config_rotation = f"{bcolors.WARNING}Disabled" if (config_allowed_settings.get('rotation') and
                                                                       not config_allowed_settings['rotation']) else \
                        f"{bcolors.OKBLUE}Enabled"
                    print(f"{bcolors.OKGREEN}\tRotation: {config_rotation}{bcolors.ENDC}")
                    print(f"{bcolors.OKGREEN}\tTunneling: {config_port_forwarding}{bcolors.ENDC}")


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: version.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rversion.proto\x12\x0fSemanticVersion\"E\n\x07Version\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\x12\r\n\x05patch\x18\x03 \x01(\x05\x12\r\n\x05\x62uild\x18\x04 \x01(\x05\x42+\n\x18\x63om.keepersecurity.protoB\x0fSemanticVersionb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'version_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\017SemanticVersion'
  _globals['_VERSION']._serialized_start=34
  _globals['_VERSION']._serialized_end=103
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: automator.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import ssocloud_pb2 as ssocloud__pb2
from . import enterprise_pb2 as enterprise__pb2
from . import version_pb2 as version__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x61utomator.proto\x12\tAutomator\x1a\x0essocloud.proto\x1a\x10\x65nterprise.proto\x1a\rversion.proto\"\xbf\x02\n\x15\x41utomatorSettingValue\x12\x11\n\tsettingId\x18\x01 \x01(\x03\x12\x15\n\rsettingTypeId\x18\x02 \x01(\x05\x12\x12\n\nsettingTag\x18\x03 \x01(\t\x12\x13\n\x0bsettingName\x18\x04 \x01(\t\x12\x14\n\x0csettingValue\x18\x05 \x01(\t\x12$\n\x08\x64\x61taType\x18\x06 \x01(\x0e\x32\x12.SsoCloud.DataType\x12\x14\n\x0clastModified\x18\x07 \x01(\t\x12\x10\n\x08\x66romFile\x18\x08 \x01(\x08\x12\x11\n\tencrypted\x18\t \x01(\x08\x12\x0f\n\x07\x65ncoded\x18\n \x01(\x08\x12\x10\n\x08\x65\x64itable\x18\x0b \x01(\x08\x12\x12\n\ntranslated\x18\x0c \x01(\x08\x12\x13\n\x0buserVisible\x18\r \x01(\x08\x12\x10\n\x08required\x18\x0e \x01(\x08\"\xee\x02\n\x14\x41pproveDeviceRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12O\n\x1dssoAuthenticationProtocolType\x18\x02 \x01(\x0e\x32(.Automator.SsoAuthenticationProtocolType\x12\x13\n\x0b\x61uthMessage\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x17\n\x0f\x64\x65vicePublicKey\x18\x05 \x01(\x0c\x12\x1c\n\x14serverEccPublicKeyId\x18\x06 \x01(\x05\x12\x1c\n\x14userEncryptedDataKey\x18\x07 \x01(\x0c\x12>\n\x18userEncryptedDataKeyType\x18\x08 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\x12\x11\n\tipAddress\x18\t \x01(\t\x12\x11\n\tisTesting\x18\n \x01(\x08\x12\x11\n\tisEccOnly\x18\x0b \x01(\x08\"\xa9\x02\n\x0cSetupRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x1c\n\x14serverEccPublicKeyId\x18\x02 \x01(\x05\x12\x31\n\x0e\x61utomatorState\x18\x03 \x01(\x0e\x32\x19.Automator.AutomatorState\x12(\n encryptedEnterprisePrivateEccKey\x18\x04 \x01(\x0c\x12(\n encryptedEnterprisePrivateRsaKey\x18\x05 \x01(\x0c\x12\x32\n\x0f\x61utomatorSkills\x18\x06 \x03(\x0b\x32\x19.Automator.AutomatorSkill\x12\x18\n\x10\x65ncryptedTreeKey\x18\x07 \x01(\x0c\x12\x11\n\tisEccOnly\x18\x08 \x01(\x08\"U\n\rStatusRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x1c\n\x14serverEccPublicKeyId\x18\x02 \x01(\x05\x12\x11\n\tisEccOnly\x18\x03 \x01(\x08\"\xa3\x04\n\x11InitializeRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x13\n\x0bidpMetadata\x18\x02 \x01(\t\x12\x1d\n\x15idpSigningCertificate\x18\x03 \x01(\x0c\x12\x13\n\x0bssoEntityId\x18\x04 \x01(\t\x12\x14\n\x0c\x65mailMapping\x18\x05 \x01(\t\x12\x18\n\x10\x66irstnameMapping\x18\x06 \x01(\t\x12\x17\n\x0flastnameMapping\x18\x07 \x01(\t\x12\x10\n\x08\x64isabled\x18\x08 \x01(\x08\x12\x1c\n\x14serverEccPublicKeyId\x18\t \x01(\x05\x12\x0e\n\x06\x63onfig\x18\n \x01(\x0c\x12\x0f\n\x07sslMode\x18\x0b \x01(\t\x12\x14\n\x0cpersistState\x18\x0c \x01(\x08\x12\x17\n\x0f\x64isableSniCheck\x18\r \x01(\x08\x12\x1e\n\x16sslCertificateFilename\x18\x0e \x01(\t\x12\"\n\x1asslCertificateFilePassword\x18\x0f \x01(\t\x12!\n\x19sslCertificateKeyPassword\x18\x10 \x01(\t\x12\x1e\n\x16sslCertificateContents\x18\x11 \x01(\x0c\x12\x15\n\rautomatorHost\x18\x12 \x01(\t\x12\x15\n\rautomatorPort\x18\x13 \x01(\t\x12\x0f\n\x07ipAllow\x18\x14 \x01(\t\x12\x0e\n\x06ipDeny\x18\x15 \x01(\t\x12\x11\n\tisEccOnly\x18\x16 \x01(\x08\"\xa6\x02\n\x16NotInitializedResponse\x12 \n\x18\x61utomatorTransmissionKey\x18\x01 \x01(\x0c\x12\x1a\n\x12signingCertificate\x18\x02 \x01(\x0c\x12\"\n\x1asigningCertificateFilename\x18\x03 \x01(\t\x12\"\n\x1asigningCertificatePassword\x18\x04 \x01(\t\x12\x1a\n\x12signingKeyPassword\x18\x05 \x01(\t\x12>\n\x18signingCertificateFormat\x18\x06 \x01(\x0e\x32\x1c.Automator.CertificateFormat\x12\x1a\n\x12\x61utomatorPublicKey\x18\x07 \x01(\x0c\x12\x0e\n\x06\x63onfig\x18\x08 \x01(\x0c\"\xa5\x04\n\x11\x41utomatorResponse\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\x12\x39\n\rapproveDevice\x18\x04 \x01(\x0b\x32 .Automator.ApproveDeviceResponseH\x00\x12+\n\x06status\x18\x05 \x01(\x0b\x32\x19.Automator.StatusResponseH\x00\x12;\n\x0enotInitialized\x18\x06 \x01(\x0b\x32!.Automator.NotInitializedResponseH\x00\x12)\n\x05\x65rror\x18\x07 \x01(\x0b\x32\x18.Automator.ErrorResponseH\x00\x12\x45\n\x13\x61pproveTeamsForUser\x18\n \x01(\x0b\x32&.Automator.ApproveTeamsForUserResponseH\x00\x12\x37\n\x0c\x61pproveTeams\x18\x0b \x01(\x0b\x32\x1f.Automator.ApproveTeamsResponseH\x00\x12\x31\n\x0e\x61utomatorState\x18\x08 \x01(\x0e\x32\x19.Automator.AutomatorState\x12\x1d\n\x15\x61utomatorPublicEccKey\x18\t \x01(\x0c\x12)\n\x07version\x18\x0c \x01(\x0b\x32\x18.SemanticVersion.VersionB\n\n\x08response\"\x98\x01\n\x15\x41pproveDeviceResponse\x12\x10\n\x08\x61pproved\x18\x01 \x01(\x08\x12\x1c\n\x14\x65ncryptedUserDataKey\x18\x02 \x01(\x0c\x12\x0f\n\x07message\x18\x03 \x01(\t\x12>\n\x18\x65ncryptedUserDataKeyType\x18\x04 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\"\xd0\x03\n\x0eStatusResponse\x12\x13\n\x0binitialized\x18\x01 \x01(\x08\x12\x18\n\x10\x65nabledTimestamp\x18\x02 \x01(\x03\x12\x1c\n\x14initializedTimestamp\x18\x03 \x01(\x03\x12\x18\n\x10updatedTimestamp\x18\x04 \x01(\x03\x12\x1f\n\x17numberOfDevicesApproved\x18\x05 \x01(\x03\x12\x1d\n\x15numberOfDevicesDenied\x18\x06 \x01(\x03\x12\x16\n\x0enumberOfErrors\x18\x07 \x01(\x03\x12 \n\x18sslCertificateExpiration\x18\x08 \x01(\x03\x12\x41\n\x16notInitializedResponse\x18\t \x01(\x0b\x32!.Automator.NotInitializedResponse\x12\x0e\n\x06\x63onfig\x18\n \x01(\x0c\x12\'\n\x1fnumberOfTeamMembershipsApproved\x18\x0b \x01(\x03\x12%\n\x1dnumberOfTeamMembershipsDenied\x18\x0c \x01(\x03\x12\x1d\n\x15numberOfTeamsApproved\x18\r \x01(\x03\x12\x1b\n\x13numberOfTeamsDenied\x18\x0e \x01(\x03\" \n\rErrorResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"X\n\x08LogEntry\x12\x12\n\nserverTime\x18\x01 \x01(\t\x12\x14\n\x0cmessageLevel\x18\x02 \x01(\t\x12\x11\n\tcomponent\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\"b\n\rAdminResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12/\n\rautomatorInfo\x18\x03 \x03(\x0b\x32\x18.Automator.AutomatorInfo\"\xee\x02\n\rAutomatorInfo\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0e\n\x06nodeId\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x04 \x01(\x08\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x32\n\x0f\x61utomatorSkills\x18\x06 \x03(\x0b\x32\x19.Automator.AutomatorSkill\x12@\n\x16\x61utomatorSettingValues\x18\x07 \x03(\x0b\x32 .Automator.AutomatorSettingValue\x12)\n\x06status\x18\x08 \x01(\x0b\x32\x19.Automator.StatusResponse\x12\'\n\nlogEntries\x18\t \x03(\x0b\x32\x13.Automator.LogEntry\x12\x31\n\x0e\x61utomatorState\x18\n \x01(\x0e\x32\x19.Automator.AutomatorState\x12\x0f\n\x07version\x18\x0b \x01(\t\"e\n\x1b\x41\x64minCreateAutomatorRequest\x12\x0e\n\x06nodeId\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12(\n\x05skill\x18\x03 \x01(\x0b\x32\x19.Automator.AutomatorSkill\"2\n\x1b\x41\x64minDeleteAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"1\n\x1f\x41\x64minGetAutomatorsOnNodeRequest\x12\x0e\n\x06nodeId\x18\x01 \x01(\x03\">\n&AdminGetAutomatorsForEnterpriseRequest\x12\x14\n\x0c\x65nterpriseId\x18\x01 \x01(\x05\"/\n\x18\x41\x64minGetAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"C\n\x1b\x41\x64minEnableAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\"\xc8\x01\n\x19\x41\x64minEditAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x03 \x01(\x08\x12\x0b\n\x03url\x18\x04 \x01(\t\x12(\n\nskillTypes\x18\x05 \x03(\x0e\x32\x14.Automator.SkillType\x12@\n\x16\x61utomatorSettingValues\x18\x06 \x03(\x0b\x32 .Automator.AutomatorSettingValue\"\xfc\x01\n\x1a\x41\x64minSetupAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x31\n\x0e\x61utomatorState\x18\x02 \x01(\x0e\x32\x19.Automator.AutomatorState\x12(\n encryptedEccEnterprisePrivateKey\x18\x03 \x01(\x0c\x12(\n encryptedRsaEnterprisePrivateKey\x18\x04 \x01(\x0c\x12(\n\nskillTypes\x18\x05 \x03(\x0e\x32\x14.Automator.SkillType\x12\x18\n\x10\x65ncryptedTreeKey\x18\x06 \x01(\x0c\"\xa6\x01\n\x1b\x41\x64minSetupAutomatorResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x13\n\x0b\x61utomatorId\x18\x03 \x01(\x03\x12\x31\n\x0e\x61utomatorState\x18\x04 \x01(\x0e\x32\x19.Automator.AutomatorState\x12\x1d\n\x15\x61utomatorEccPublicKey\x18\x05 \x01(\x0c\"2\n\x1b\x41\x64minAutomatorSkillsRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"_\n\x0e\x41utomatorSkill\x12\'\n\tskillType\x18\x01 \x01(\x0e\x32\x14.Automator.SkillType\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x16\n\x0etranslatedName\x18\x03 \x01(\t\"t\n\x1c\x41\x64minAutomatorSkillsResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x32\n\x0f\x61utomatorSkills\x18\x03 \x03(\x0b\x32\x19.Automator.AutomatorSkill\"1\n\x1a\x41\x64minResetAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"6\n\x1f\x41\x64minInitializeAutomatorRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"/\n\x18\x41\x64minAutomatorLogRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"4\n\x1d\x41\x64minAutomatorLogClearRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\"\xe3\x02\n\x1a\x41pproveTeamsForUserRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12O\n\x1dssoAuthenticationProtocolType\x18\x02 \x01(\x0e\x32(.Automator.SsoAuthenticationProtocolType\x12\x13\n\x0b\x61uthMessage\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x1c\n\x14serverEccPublicKeyId\x18\x05 \x01(\x05\x12\x11\n\tipAddress\x18\x06 \x01(\t\x12\x15\n\ruserPublicKey\x18\x07 \x01(\x0c\x12\x33\n\x0fteamDescription\x18\x08 \x03(\x0b\x32\x1a.Automator.TeamDescription\x12\x11\n\tisTesting\x18\t \x01(\x08\x12\x11\n\tisEccOnly\x18\n \x01(\x08\x12\x18\n\x10userPublicKeyEcc\x18\x0b \x01(\x0c\"\x8a\x01\n\x0fTeamDescription\x12\x0f\n\x07teamUid\x18\x01 \x01(\x0c\x12\x10\n\x08teamName\x18\x02 \x01(\t\x12\x18\n\x10\x65ncryptedTeamKey\x18\x03 \x01(\x0c\x12:\n\x14\x65ncryptedTeamKeyType\x18\x04 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\"\x99\x01\n\x1b\x41pproveTeamsForUserResponse\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x45\n\x13\x61pproveTeamResponse\x18\x04 \x03(\x0b\x32(.Automator.ApproveOneTeamForUserResponse\"\xab\x02\n\x1d\x41pproveOneTeamForUserResponse\x12\x10\n\x08\x61pproved\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0f\n\x07teamUid\x18\x03 \x01(\x0c\x12\x10\n\x08teamName\x18\x04 \x01(\t\x12\x1c\n\x14userEncryptedTeamKey\x18\x05 \x01(\x0c\x12>\n\x18userEncryptedTeamKeyType\x18\x06 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\x12!\n\x19userEncryptedTeamKeyByEcc\x18\x07 \x01(\x0c\x12\x43\n\x1duserEncryptedTeamKeyByEccType\x18\x08 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\"\xab\x02\n\x13\x41pproveTeamsRequest\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12O\n\x1dssoAuthenticationProtocolType\x18\x02 \x01(\x0e\x32(.Automator.SsoAuthenticationProtocolType\x12\x13\n\x0b\x61uthMessage\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x1c\n\x14serverEccPublicKeyId\x18\x05 \x01(\x05\x12\x11\n\tipAddress\x18\x06 \x01(\t\x12\x33\n\x0fteamDescription\x18\x07 \x03(\x0b\x32\x1a.Automator.TeamDescription\x12\x11\n\tisEccOnly\x18\x08 \x01(\x08\x12\x11\n\tisTesting\x18\t \x01(\x08\"|\n\x14\x41pproveTeamsResponse\x12\x13\n\x0b\x61utomatorId\x18\x01 \x01(\x03\x12\x0f\n\x07message\x18\x02 \x01(\t\x12>\n\x13\x61pproveTeamResponse\x18\x03 \x03(\x0b\x32!.Automator.ApproveOneTeamResponse\"\x9e\x04\n\x16\x41pproveOneTeamResponse\x12\x10\n\x08\x61pproved\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0f\n\x07teamUid\x18\x03 \x01(\x0c\x12\x10\n\x08teamName\x18\x04 \x01(\t\x12\x1b\n\x13\x65ncryptedTeamKeyCbc\x18\x05 \x01(\x0c\x12=\n\x17\x65ncryptedTeamKeyCbcType\x18\x06 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\x12\x1b\n\x13\x65ncryptedTeamKeyGcm\x18\x07 \x01(\x0c\x12=\n\x17\x65ncryptedTeamKeyGcmType\x18\x08 \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\x12\x18\n\x10teamPublicKeyRsa\x18\t \x01(\x0c\x12\"\n\x1a\x65ncryptedTeamPrivateKeyRsa\x18\n \x01(\x0c\x12\x44\n\x1e\x65ncryptedTeamPrivateKeyRsaType\x18\x0b \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType\x12\x18\n\x10teamPublicKeyEcc\x18\x0c \x01(\x0c\x12\"\n\x1a\x65ncryptedTeamPrivateKeyEcc\x18\r \x01(\x0c\x12\x44\n\x1e\x65ncryptedTeamPrivateKeyEccType\x18\x0e \x01(\x0e\x32\x1c.Enterprise.EncryptedKeyType*@\n\x1dSsoAuthenticationProtocolType\x12\x14\n\x10UNKNOWN_PROTOCOL\x10\x00\x12\t\n\x05SAML2\x10\x01*<\n\x11\x43\x65rtificateFormat\x12\x12\n\x0eUNKNOWN_FORMAT\x10\x00\x12\n\n\x06PKCS12\x10\x01\x12\x07\n\x03JKS\x10\x02*g\n\tSkillType\x12\x16\n\x12UNKNOWN_SKILL_TYPE\x10\x00\x12\x13\n\x0f\x44\x45VICE_APPROVAL\x10\x01\x12\x11\n\rTEAM_APPROVAL\x10\x02\x12\x1a\n\x16TEAM_FOR_USER_APPROVAL\x10\x03*\x87\x01\n\x0e\x41utomatorState\x12\x11\n\rUNKNOWN_STATE\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\t\n\x05\x45RROR\x10\x02\x12\x18\n\x14NEEDS_INITIALIZATION\x10\x03\x12\x17\n\x13NEEDS_CRYPTO_STEP_1\x10\x04\x12\x17\n\x13NEEDS_CRYPTO_STEP_2\x10\x05\x42%\n\x18\x63om.keepersecurity.protoB\tAutomatorb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'automator_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\tAutomator'
  _globals['_SSOAUTHENTICATIONPROTOCOLTYPE']._serialized_start=7193
  _globals['_SSOAUTHENTICATIONPROTOCOLTYPE']._serialized_end=7257
  _globals['_CERTIFICATEFORMAT']._serialized_start=7259
  _globals['_CERTIFICATEFORMAT']._serialized_end=7319
  _globals['_SKILLTYPE']._serialized_start=7321
  _globals['_SKILLTYPE']._serialized_end=7424
  _globals['_AUTOMATORSTATE']._serialized_start=7427
  _globals['_AUTOMATORSTATE']._serialized_end=7562
  _globals['_AUTOMATORSETTINGVALUE']._serialized_start=80
  _globals['_AUTOMATORSETTINGVALUE']._serialized_end=399
  _globals['_APPROVEDEVICEREQUEST']._serialized_start=402
  _globals['_APPROVEDEVICEREQUEST']._serialized_end=768
  _globals['_SETUPREQUEST']._serialized_start=771
  _globals['_SETUPREQUEST']._serialized_end=1068
  _globals['_STATUSREQUEST']._serialized_start=1070
  _globals['_STATUSREQUEST']._serialized_end=1155
  _globals['_INITIALIZEREQUEST']._serialized_start=1158
  _globals['_INITIALIZEREQUEST']._serialized_end=1705
  _globals['_NOTINITIALIZEDRESPONSE']._serialized_start=1708
  _globals['_NOTINITIALIZEDRESPONSE']._serialized_end=2002
  _globals['_AUTOMATORRESPONSE']._serialized_start=2005
  _globals['_AUTOMATORRESPONSE']._serialized_end=2554
  _globals['_APPROVEDEVICERESPONSE']._serialized_start=2557
  _globals['_APPROVEDEVICERESPONSE']._serialized_end=2709
  _globals['_STATUSRESPONSE']._serialized_start=2712
  _globals['_STATUSRESPONSE']._serialized_end=3176
  _globals['_ERRORRESPONSE']._serialized_start=3178
  _globals['_ERRORRESPONSE']._serialized_end=3210
  _globals['_LOGENTRY']._serialized_start=3212
  _globals['_LOGENTRY']._serialized_end=3300
  _globals['_ADMINRESPONSE']._serialized_start=3302
  _globals['_ADMINRESPONSE']._serialized_end=3400
  _globals['_AUTOMATORINFO']._serialized_start=3403
  _globals['_AUTOMATORINFO']._serialized_end=3769
  _globals['_ADMINCREATEAUTOMATORREQUEST']._serialized_start=3771
  _globals['_ADMINCREATEAUTOMATORREQUEST']._serialized_end=3872
  _globals['_ADMINDELETEAUTOMATORREQUEST']._serialized_start=3874
  _globals['_ADMINDELETEAUTOMATORREQUEST']._serialized_end=3924
  _globals['_ADMINGETAUTOMATORSONNODEREQUEST']._serialized_start=3926
  _globals['_ADMINGETAUTOMATORSONNODEREQUEST']._serialized_end=3975
  _globals['_ADMINGETAUTOMATORSFORENTERPRISEREQUEST']._serialized_start=3977
  _globals['_ADMINGETAUTOMATORSFORENTERPRISEREQUEST']._serialized_end=4039
  _globals['_ADMINGETAUTOMATORREQUEST']._serialized_start=4041
  _globals['_ADMINGETAUTOMATORREQUEST']._serialized_end=4088
  _globals['_ADMINENABLEAUTOMATORREQUEST']._serialized_start=4090
  _globals['_ADMINENABLEAUTOMATORREQUEST']._serialized_end=4157
  _globals['_ADMINEDITAUTOMATORREQUEST']._serialized_start=4160
  _globals['_ADMINEDITAUTOMATORREQUEST']._serialized_end=4360
  _globals['_ADMINSETUPAUTOMATORREQUEST']._serialized_start=4363
  _globals['_ADMINSETUPAUTOMATORREQUEST']._serialized_end=4615
  _globals['_ADMINSETUPAUTOMATORRESPONSE']._serialized_start=4618
  _globals['_ADMINSETUPAUTOMATORRESPONSE']._serialized_end=4784
  _globals['_ADMINAUTOMATORSKILLSREQUEST']._serialized_start=4786
  _globals['_ADMINAUTOMATORSKILLSREQUEST']._serialized_end=4836
  _globals['_AUTOMATORSKILL']._serialized_start=4838
  _globals['_AUTOMATORSKILL']._serialized_end=4933
  _globals['_ADMINAUTOMATORSKILLSRESPONSE']._serialized_start=4935
  _globals['_ADMINAUTOMATORSKILLSRESPONSE']._serialized_end=5051
  _globals['_ADMINRESETAUTOMATORREQUEST']._serialized_start=5053
  _globals['_ADMINRESETAUTOMATORREQUEST']._serialized_end=5102
  _globals['_ADMININITIALIZEAUTOMATORREQUEST']._serialized_start=5104
  _globals['_ADMININITIALIZEAUTOMATORREQUEST']._serialized_end=5158
  _globals['_ADMINAUTOMATORLOGREQUEST']._serialized_start=5160
  _globals['_ADMINAUTOMATORLOGREQUEST']._serialized_end=5207
  _globals['_ADMINAUTOMATORLOGCLEARREQUEST']._serialized_start=5209
  _globals['_ADMINAUTOMATORLOGCLEARREQUEST']._serialized_end=5261
  _globals['_APPROVETEAMSFORUSERREQUEST']._serialized_start=5264
  _globals['_APPROVETEAMSFORUSERREQUEST']._serialized_end=5619
  _globals['_TEAMDESCRIPTION']._serialized_start=5622
  _globals['_TEAMDESCRIPTION']._serialized_end=5760
  _globals['_APPROVETEAMSFORUSERRESPONSE']._serialized_start=5763
  _globals['_APPROVETEAMSFORUSERRESPONSE']._serialized_end=5916
  _globals['_APPROVEONETEAMFORUSERRESPONSE']._serialized_start=5919
  _globals['_APPROVEONETEAMFORUSERRESPONSE']._serialized_end=6218
  _globals['_APPROVETEAMSREQUEST']._serialized_start=6221
  _globals['_APPROVETEAMSREQUEST']._serialized_end=6520
  _globals['_APPROVETEAMSRESPONSE']._serialized_start=6522
  _globals['_APPROVETEAMSRESPONSE']._serialized_end=6646
  _globals['_APPROVEONETEAMRESPONSE']._serialized_start=6649
  _globals['_APPROVEONETEAMRESPONSE']._serialized_end=7191
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import datetime
import json
import logging

from .base import GroupCommand, Command, report_output_parser, dump_report_data, field_to_title, user_choice
from .. import api, utils
from ..error import CommandError, KeeperApiError
from ..proto import APIRequest_pb2


class TwoFaCommand(GroupCommand):
    def __init__(self):
        super().__init__()
        self.register_command('list', TfaListCommand(), 'Displays a list of 2FA methods')
        self.register_command('add', TfaAddCommand(), 'Add 2FA method')
        self.register_command('delete', TfaDeleteCommand(), 'Delete 2FA method')
        self.default_verb = 'list'

    @staticmethod
    def two_factor_channel_to_desc(channel):
        if channel == APIRequest_pb2.TWO_FA_CT_TOTP:
            return 'TOTP'
        if channel == APIRequest_pb2.TWO_FA_CT_SMS:
            return 'SMS'
        if channel == APIRequest_pb2.TWO_FA_CT_DUO:
            return 'DUO'
        if channel == APIRequest_pb2.TWO_FA_CT_RSA:
            return 'RSA SecurID'
        if channel == APIRequest_pb2.TWO_FA_CT_U2F:
            return 'U2F'
        if channel == APIRequest_pb2.TWO_FA_CT_WEBAUTHN:
            return 'Security Key'
        if channel == APIRequest_pb2.TWO_FA_CT_DNA:
            return 'Keeper DNA (Watch)'
        if channel == APIRequest_pb2.TWO_FA_CT_BACKUP:
            return 'Backup Codes'


class TfaListCommand(Command):
    parser = argparse.ArgumentParser(prog='2fa list', parents=[report_output_parser])

    def get_parser(self):
        return TfaListCommand.parser

    def execute(self, params, **kwargs):
        rs = api.communicate_rest(params, None, 'authentication/2fa_list', rs_type=APIRequest_pb2.TwoFactorListResponse)
        expire_at = ''
        if rs.expireOn > 0:
            if rs.expireOn > 3_000_000_000_000:
                expire_at = 'Never'
            else:
                dt = datetime.datetime.fromtimestamp(rs.expireOn // 1000)
                expire_at = dt.isoformat()

        if expire_at is not None:
            logging.info('2FA authentication expires: %s\n', expire_at)
        table = []
        for channel in rs.channels:
            created_on = datetime.datetime.fromtimestamp(channel.createdOn // 1000)
            row = [TwoFaCommand.two_factor_channel_to_desc(channel.channelType),
                   utils.base64_url_encode(channel.channel_uid), channel.channelName, created_on, channel.phoneNumber]
            table.append(row)

        fmt = kwargs.get('format')
        header = ['method', 'channel_uid', 'name', 'created', 'phone_number']
        if fmt != 'json':
            header = [field_to_title(x) for x in header]

        return dump_report_data(table, header, fmt=fmt, filename=kwargs.get('output'), row_number=True)


class TfaAddCommand(Command):
    parser = argparse.ArgumentParser(prog='2fa add', description='Add 2FA method')
    parser.add_argument('--method', '-m', dest='method', action='store', required=True,
                        choices=['totp', 'key', 'sms', 'duo', 'backup'], help='2FA auth method')
    parser.add_argument('--name', dest='name', action='store', help='2FA auth name')
    parser.add_argument('--key-pin', dest='key_pin', action='store_true', help='force using Security Key PIN')

    def get_parser(self):
        return TfaAddCommand.parser

    def execute(self, params, **kwargs):
        all_tfa_restrictions = {'require_security_key_pin', 'restrict_two_factor_channel_text',
                                'restrict_two_factor_channel_google', 'restrict_two_factor_channel_duo',
                                'restrict_two_factor_channel_security_key', 'restrict_two_factor_channel_rsa',
                                'restrict_two_factor_channel_dna'}
        tfa_restrictions = set()
        if params.enforcements and 'booleans' in params.enforcements:
            for x in params.enforcements['booleans']:
                key = (x.get('key') or '').lower()
                if key in all_tfa_restrictions:
                    tfa_restrictions.add(key)

        method = kwargs.get('method')
        rq = APIRequest_pb2.TwoFactorAddRequest()
        rq.channel_uid = utils.base64_url_decode(utils.generate_uid())
        rq.channelName = kwargs.get('name') or ''

        if method == 'totp':
            if 'restrict_two_factor_channel_google' in tfa_restrictions:
                raise CommandError('', 'Authenticator App (TOTP) 2FA method is disabled by the Administrator')
            rq.channelType = APIRequest_pb2.TWO_FA_CT_TOTP
        elif method == 'sms':
            if 'restrict_two_factor_channel_text' in tfa_restrictions:
                raise CommandError('', 'Text Message (SMS) 2FA method is disabled by the Administrator')
            rq.channelType = APIRequest_pb2.TWO_FA_CT_SMS
        elif method == 'key':
            if 'restrict_two_factor_channel_security_key' in tfa_restrictions:
                raise CommandError('', 'Security Key 2FA method is disabled by the Administrator')
            rq.channelType = APIRequest_pb2.TWO_FA_CT_WEBAUTHN
        elif method == 'duo':
            rq.channelType = APIRequest_pb2.TWO_FA_CT_DUO
        elif method == 'backup':
            rq.channelType = APIRequest_pb2.TWO_FA_CT_BACKUP
        else:
            raise CommandError('2fa add', f'2FA method "{method}" is not supported')

        # Input
        if rq.channelType == APIRequest_pb2.TWO_FA_CT_SMS:
            try:
                phone_number = input('\nEnter your phone number for text messages: ')
                if not phone_number:
                    return
                rq.phoneNumber = phone_number
            except KeyboardInterrupt:
                return

        elif rq.channelType == APIRequest_pb2.TWO_FA_CT_DUO:
            duo_rs = api.communicate_rest(params, None, 'authentication/2fa_duo_status',
                                          rs_type=APIRequest_pb2.TwoFactorDuoStatus)
            if duo_rs.enroll_url:
                logging.warning(duo_rs.message)
                logging.warning("Enroll URL")
                print(duo_rs.enroll_url)
                return
            capabilities = [x for x in duo_rs.capabilities if x in ('mobile_otp', 'sms', 'voice')]
            print(f'Device Phone Number: {duo_rs.phoneNumber}')
            if len(capabilities) == 0:
                return
            print('We\'ll send you a text message or call with a passcode to your device:')
            for no in range(len(capabilities)):
                c = capabilities[no]
                if c == 'sms':
                    c = 'Send a Text Message'
                elif c == 'voice':
                    c = 'Make a Voice Call'
                if c == 'mobile_otp':
                    c = 'OTP Code on Mobile'
                print(f'  {no+1}. {c}')
            print(f'  q. Cancel')
            while True:
                answer = input('Selection: ')
                if answer in ('q', 'Q'):
                    return
                if answer:
                    if answer.isnumeric():
                        code = int(answer)
                        if 0 < code < len(capabilities):
                            c = capabilities[code - 1]
                            rq.duoPushType = \
                                APIRequest_pb2.TWO_FA_PUSH_DUO_TEXT if c == 'sms' else \
                                APIRequest_pb2.TWO_FA_PUSH_DUO_CALL if c == 'voice' else \
                                APIRequest_pb2.TWO_FA_PUSH_NONE
                            break
                    logging.info(f'Action \"{answer}\" is not supported.')

        rs = api.communicate_rest(params, rq, 'authentication/2fa_add', rs_type=APIRequest_pb2.TwoFactorAddResponse)

        if rq.channelType == APIRequest_pb2.TWO_FA_CT_BACKUP:
            codes = list(rs.backupKeys)
            table = []
            for no in range(0, len(codes), 2):
                table.append(codes[no: no+2])
            dump_report_data(table, ('', ''), title='Backup Codes', no_header=True)
            return

        if rq.channelType == APIRequest_pb2.TWO_FA_CT_WEBAUTHN:
            try:
                from ..yubikey.yubikey import yubikey_register
                request = json.loads(rs.challenge)
                force_pin = kwargs.get('key_pin') is True
                response = yubikey_register(request, force_pin)
                credential_id = response.id
                attestation = {
                    'id': credential_id,
                    'rawId': utils.base64_url_encode(response.raw_id),
                    'response': {
                        'attestationObject': utils.base64_url_encode(response.response.attestation_object),
                        'clientDataJSON': response.response.client_data.b64
                    },
                    'type': 'public-key',
                    # 'transports': ['usb'],
                    'clientExtensionResults': response.client_extension_results or {}
                }
                rq_yubikey = APIRequest_pb2.TwoFactorValidateRequest()
                rq_yubikey.valueType = APIRequest_pb2.TWO_FA_RESP_WEBAUTHN
                rq_yubikey.value = json.dumps(attestation)
                rq_yubikey.channel_uid = rq.channel_uid
                rq_yubikey.expireIn = APIRequest_pb2.TWO_FA_EXP_IMMEDIATELY
                api.communicate_rest(params, rq_yubikey, 'authentication/2fa_add_validate')
            except ImportError as e:
                from ..yubikey import display_fido2_warning
                logging.warning(e)
                display_fido2_warning()
            except Exception as e:
                logging.warning(e)
            return

        if rq.channelType == APIRequest_pb2.TWO_FA_CT_TOTP:
            url = f'otpauth://totp/Keeper:{params.user}?secret={rs.challenge}'
            print(f'TOTP URL:\n{url}')
            try:
                import pyqrcode
                url = pyqrcode.create(url)
                print(url.terminal('black', 'white'))
            except ModuleNotFoundError:
                print('QR Code library is not installed.\npip install pyqrcode')

        rq_validate = APIRequest_pb2.TwoFactorValidateRequest()
        if rq.channelType == APIRequest_pb2.TWO_FA_CT_TOTP:
            rq_validate.valueType = APIRequest_pb2.TWO_FA_CODE_TOTP
        elif rq.channelType == APIRequest_pb2.TWO_FA_CT_SMS:
            rq_validate.valueType = APIRequest_pb2.TWO_FA_CODE_SMS
        elif rq.channelType == APIRequest_pb2.TWO_FA_CT_WEBAUTHN:
            rq_validate.valueType = APIRequest_pb2.TWO_FA_RESP_WEBAUTHN
        elif rq.channelType == APIRequest_pb2.TWO_FA_CT_DUO:
            rq_validate.valueType = APIRequest_pb2.TWO_FA_CODE_DUO

        rq_validate.expireIn = APIRequest_pb2.TWO_FA_EXP_IMMEDIATELY
        rq_validate.channel_uid = rq.channel_uid
        while True:
            try:
                answer = input('Verification Code: ')
                if answer:
                    rq_validate.value = answer
                    try:
                        api.communicate_rest(params, rq_validate, 'authentication/2fa_add_validate')
                        print("\033[2J")
                        logging.info('2FA method is added')
                        return
                    except KeeperApiError as kae:
                        logging.warning('Invalid 2FA code: (%s): %s ', kae.result_code, kae.message)
            except KeyboardInterrupt:
                return


class TfaDeleteCommand(Command):
    parser = argparse.ArgumentParser(prog='2fa delete', description='Delete 2FA method')
    parser.add_argument('--force', dest='force', action='store_true', help='do not prompt for confirmation')
    parser.add_argument('name', help='2FA method UID or name')

    def get_parser(self):
        return TfaDeleteCommand.parser

    def execute(self, params, **kwargs):
        name = kwargs.get('name')
        if not name:
            raise CommandError('2fa delete', '"name" argument is required')
        rs = api.communicate_rest(params, None, 'authentication/2fa_list', rs_type=APIRequest_pb2.TwoFactorListResponse)
        channel = next((x for x in rs.channels if utils.base64_url_encode(x.channel_uid) == name), None)
        if not channel:
            l_name = name.casefold()
            for x in rs.channels:
                if l_name == x.channelName.casefold():
                    channel = x
                    break
        if not channel:
            raise CommandError('2fa delete', f'2FA channel "{name}" not found')

        if not kwargs.get('force') is True:
            channel_name = channel.channelName
            if not channel_name:
                channel_name = utils.base64_url_encode(channel.channel_uid)
            answer = user_choice(f'Do you want to delete 2FA channel "{channel_name}"?', 'yn', 'n')
            if answer not in ('y', 'Y'):
                return
        rq = APIRequest_pb2.TwoFactorDeleteRequest()
        rq.channel_uid = channel.channel_uid
        api.communicate_rest(params, rq, 'authentication/2fa_delete')
        logging.info('2FA channel is deleted')



from __future__ import annotations
import argparse
from ..discover import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ... import vault
from ...discovery_common.constants import PAM_USER, PAM_MACHINE
from ...discovery_common.user_service import UserService
from ...display import bcolors
from ... import __version__
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ...vault import TypedRecord
    from ...params import KeeperParams


class PAMActionServiceRemoveCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-service-remove')

    # The record to base everything on.
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name or UID')

    parser.add_argument('--machine-uid', '-m', required=True, dest='machine_uid', action='store',
                        help='The UID of the Windows Machine record')
    parser.add_argument('--user-uid', '-u', required=True, dest='user_uid', action='store',
                        help='The UID of the User record')
    parser.add_argument('--type', '-t', required=True, choices=['service', 'task', 'iis'], dest='type',
                        action='store', help='Relationship to remove [service, task, iis]')

    def get_parser(self):
        return PAMActionServiceRemoveCommand.parser

    def execute(self, params: KeeperParams, **kwargs):

        gateway = kwargs.get("gateway")
        machine_uid = kwargs.get("machine_uid")
        user_uid = kwargs.get("user_uid")
        rel_type = kwargs.get("type")

        print("")

        gateway_context = GatewayContext.from_gateway(params, gateway)
        if gateway_context is None:
            print(f"{bcolors.FAIL}Could not find the gateway configuration for {gateway}.")
            return

        if gateway_context is None:
            print(f"  {self._f('Cannot get gateway information. Gateway may not be up.')}")
            return

        user_service = UserService(record=gateway_context.configuration, params=params, fail_on_corrupt=False,
                                   agent=f"Cmdr/{__version__}")

        machine_record = vault.KeeperRecord.load(params, machine_uid)  # type: Optional[TypedRecord]
        if machine_record is None:
            print(self._f("The machine record does not exists."))
            return

        if machine_record.record_type != PAM_MACHINE:
            print(self._f("The machine record is not a PAM Machine."))
            return

        user_record = vault.KeeperRecord.load(params, user_uid)  # type: Optional[TypedRecord]
        if user_record is None:
            print(self._f("The user record does not exists."))
            return

        if user_record.record_type != PAM_USER:
            print(self._f("The user record is not a PAM User."))
            return

        machine_vertex = user_service.get_record_link(machine_record.record_uid)
        if machine_vertex is None:
            print(self._f(f"The machine does not exist in the mapping."))
            return

        user_vertex = user_service.get_record_link(user_record.record_uid)
        if user_vertex is None:
            print(self._f(f"The user does not exist in the mapping."))
            return

        acl = user_service.get_acl(machine_vertex.uid, user_vertex.uid)
        if acl is None:
            print(f"{bcolors.WARNING}The user did not control any services, "
                  f"scheduled tasks, or IIS pools on the machine.{bcolors.ENDC}")
            return

        if rel_type == "service":
            acl.is_service = False
        elif rel_type == "task":
            acl.is_task = False
        else:
            acl.is_iis_pool = False

        if user_service.dag.get_root.has(machine_vertex) is False:
            user_service.belongs_to(gateway_context.configuration_uid, machine_vertex.uid)

        user_service.belongs_to(machine_vertex.uid, user_vertex.uid, acl=acl)
        user_service.save()

        if rel_type == "service":
            print(
                self._gr(
                    "Success: Services running on this machine will no longer have their password changed when this "
                    "user's password is rotated."
                )
            )
        elif rel_type == "task":
            print(
                self._gr(
                    "Success: Scheduled tasks running on this machine will no longer have their password changed "
                    "when this user's password is rotated."
                )
            )
        else:
            print(
                self._gr(
                    "Success: IIP pools running on this machine will no longer have their password changed "
                    "when this user's password is rotated."
                )
            )



import sys
import unittest

if sys.version_info >= (3, 8):
    import pytest
    from unittest import TestCase, mock
    from flask import Flask
    from keepercommander.service.util.command_util import CommandExecutor
    from keepercommander.service.util.exceptions import CommandExecutionError
    from keepercommander.service.util.parse_keeper_response import parse_keeper_response

    class TestCommandAPI(TestCase):
        def setUp(self):
            self.app = Flask(__name__)
            self.client = self.app.test_client()
            
            @self.app.route('/api/v1/executecommand', methods=['POST'])
            def execute_command():
                command = "ls"
                response, status_code = CommandExecutor.execute(command)
                return {'response': response}, status_code

        def test_validate_command(self):
            """Test command validation"""
            result, status_code = CommandExecutor.validate_command("")
            self.assertIsNotNone(result)
            self.assertEqual(status_code, 400)
            self.assertEqual(result["error"], "No command provided.")

            result = CommandExecutor.validate_command("ls")
            self.assertIsNone(result)

        def test_validate_session(self):
            """Test session validation"""
            with mock.patch('keepercommander.service.util.command_util.get_current_params', return_value=None):
                result, status_code = CommandExecutor.validate_session()
                self.assertEqual(status_code, 401)
                self.assertIn("No active session", result["error"])

            with mock.patch('keepercommander.service.util.command_util.get_current_params', return_value={"session": "active"}):
                result = CommandExecutor.validate_session()
                self.assertIsNone(result)

        @unittest.skip
        def test_command_execution_success(self):
            """Test successful command execution"""
            mock_params = {"session": "active"}
            test_command = "ls"
            expected_output = "Folder1\nFolder2\n"

            with mock.patch('keepercommander.service.util.command_util.get_current_params', return_value=mock_params), \
                mock.patch('keepercommander.cli.do_command', return_value=expected_output), \
                mock.patch('keepercommander.service.util.command_util.ConfigReader.read_config', return_value=None):

                response, status_code = CommandExecutor.execute(test_command)
                self.assertEqual(status_code, 200)
                self.assertIsNotNone(response)

        @unittest.skip
        def test_command_execution_failure(self):
            """Test command execution failure"""
            mock_params = {"session": "active"}
            test_command = "invalid_command"

            with mock.patch('keepercommander.service.util.command_util.get_current_params', return_value=mock_params), \
                mock.patch('keepercommander.cli.do_command', side_effect=Exception("Command failed")), \
                self.assertRaises(CommandExecutionError):
                
                CommandExecutor.execute(test_command)

        def test_response_encryption(self):
            """Test response encryption when key is present"""
            test_response = {"status": "success", "data": "test"}

            mock_key = "0" * 32

            with mock.patch('keepercommander.service.util.command_util.ConfigReader.read_config', return_value=mock_key):
                encrypted_response = CommandExecutor.encrypt_response(test_response)
                self.assertIsInstance(encrypted_response, bytes)
                self.assertGreater(len(encrypted_response), 0)

        def test_response_parsing(self):
            """Test response parsing for different commands"""

            ls_response = "# Folder UID\n1 folder1_uid folder1 rw\n# Record UID\n1 record1_uid login record1"
            parsed = parse_keeper_response("ls", ls_response)
            self.assertEqual(parsed["status"], "success")
            self.assertEqual(parsed["command"], "ls")
            self.assertIn("folders", parsed["data"])
            self.assertIn("records", parsed["data"])

            tree_response = "Root\n  Folder1\n    SubFolder1"
            parsed = parse_keeper_response("tree", tree_response)
            self.assertEqual(parsed["command"], "tree")
            self.assertIsInstance(parsed["data"], dict)
            self.assertIn("tree", parsed["data"])

        def test_capture_output(self):
            """Test command output capture"""
            test_command = "ls"
            expected_output = "test output"
            mock_params = {"session": "active"}

            with mock.patch('keepercommander.cli.do_command', return_value=expected_output):
                return_value, output, logs = CommandExecutor.capture_output_and_logs(mock_params, test_command)
                self.assertEqual(return_value, expected_output)

        @unittest.skip
        def test_integration_command_flow(self):
            """Test the complete command execution flow"""
            test_command = "ls"
            mock_params = {"session": "active"}
            expected_output = "# Folder UID\n1 folder1_uid folder1 rw"

            with mock.patch('keepercommander.service.util.command_util.get_current_params', return_value=mock_params), \
                mock.patch('keepercommander.cli.do_command', return_value=expected_output), \
                mock.patch('keepercommander.service.util.command_util.ConfigReader.read_config', return_value=None):

                response, status_code = CommandExecutor.execute(test_command)
                self.assertEqual(status_code, 200)
                self.assertIsNotNone(response)


from __future__ import annotations
import os
from .__version__ import __version__
from .constants import PAM_USER
from .types import DiscoveryObject
from ..keeper_dag.vertex import DAGVertex
from typing import List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..keeper_dag.dag import DAG


def value_to_boolean(value):
    value = str(value)
    if value.lower() in ['true', 'yes', 'on', '1']:
        return True
    elif value.lower() in ['false', 'no', 'off', '0']:
        return False
    else:
        return None


def get_connection(**kwargs):

    """
    This method will return the proper connection based on the params passed in.

    If `ksm` and a KDNRM KSM instance, it will connect using keeper secret manager.
    If `params` and a KeeperParam instance, it will connect using Commander.
    If the env var `USE_LOCAL_DAG` is True, it will connect using the Local test DAG engine.

    It returns a child instance of the Connection class.
    """

    # if the connection is passed in, return it.
    if kwargs.get("connection") is not None:
        return kwargs.get("connection")

    ksm = kwargs.get("ksm")
    params = kwargs.get("params")
    logger = kwargs.get("logger")
    if value_to_boolean(os.environ.get("USE_LOCAL_DAG")):
        from ..keeper_dag.connection.local import Connection
        conn = Connection(logger=logger)
    else:
        if ksm is not None:
            from ..keeper_dag.connection.ksm import Connection
            conn = Connection(config=ksm.storage_config, logger=logger)
        elif params is not None:
            from ..keeper_dag.connection.commander import Connection
            conn = Connection(params=params, logger=logger)
        else:
            raise ValueError("Must pass 'ksm' for KSM, 'params' for Commander. Found neither.")
    return conn


def split_user_and_domain(user: str) -> (Optional[str], Optional[str]):

    if user is None:
        return None, None

    domain = None

    if "\\" in user:
        user_parts = user.split("\\", maxsplit=1)
        user = user_parts[0]
        domain = user_parts[1]
    elif "@" in user:
        user_parts = user.split("@")
        domain = user_parts.pop()
        user = "@".join(user_parts)

    return user, domain


def user_check_list(user: str, name: Optional[str] = None, source: Optional[str] = None) -> List[str]:
    user, domain = split_user_and_domain(user)
    user = user.lower()
    check_list = [user, f".\\{user}", ]
    if name is not None:
        name = name.lower()
        check_list += [name, f".\\{name}"]
    if source is not None:
        check_list.append(f"{source.lower()[:15]}\\{user}")
        check_list.append(f"{user}@{source.lower()}")
        netbios_parts = source.split(".")
        if len(netbios_parts) > 1:
            check_list.append(f"{netbios_parts[0][:15]}\\{user}")
            check_list.append(f"{user}@{netbios_parts[0]}")
    if domain is not None:
        domain = domain.lower()
        check_list.append(f"{domain}[:15]\\{user}")
        check_list.append(f"{user}@{domain}")
        domain_parts = domain.split(".")
        if len(domain_parts) > 1:
            check_list.append(f"{domain_parts[0][:15]}\\{user}")
            check_list.append(f"\\{user}@{domain_parts[0]}")

    return list(set(check_list))


def user_in_lookup(user: str, lookup: dict, name: Optional[str] = None, source: Optional[str] = None) -> bool:

    for check_user in user_check_list(user, name, source):
        if check_user in lookup:
            return True
    return False


def find_user_vertex(graph: DAG, user: str, domain: Optional[str] = None) -> Optional[DAGVertex]:

    user_vertices = graph.search_content({"record_type": PAM_USER})
    for user_vertex in user_vertices:

        # Make sure the vertex is active, and has content data
        if user_vertex.active is False or user_vertex.has_data is False:
            continue
        content = DiscoveryObject.get_discovery_object(user_vertex)

        current_user, current_domain = split_user_and_domain(content.item.user)

        # If we are want a directory user and the current user is not one, or does not match the domain, then skip
        if domain is not None and (current_domain is None or domain.lower() != current_domain.lower()):
            continue

        if current_user.lower() == user.lower():
            return user_vertex

    return None


def make_agent(text) -> str:
    return f"{text}/{__version__}"



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Version(_message.Message):
    __slots__ = ["major", "minor", "patch", "build"]
    MAJOR_FIELD_NUMBER: _ClassVar[int]
    MINOR_FIELD_NUMBER: _ClassVar[int]
    PATCH_FIELD_NUMBER: _ClassVar[int]
    BUILD_FIELD_NUMBER: _ClassVar[int]
    major: int
    minor: int
    patch: int
    build: int
    def __init__(self, major: _Optional[int] = ..., minor: _Optional[int] = ..., patch: _Optional[int] = ..., build: _Optional[int] = ...) -> None: ...
